% This file is part of the Attempto Parsing Engine (APE).
% Copyright 2008, Attempto Group, University of Zurich (see http://attempto.ifi.uzh.ch).
%
% The Attempto Parsing Engine (APE) is free software: you can redistribute it and/or modify it
% under the terms of the GNU Lesser General Public License as published by the Free Software
% Foundation, either version 3 of the License, or (at your option) any later version.
%
% The Attempto Parsing Engine (APE) is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
% PURPOSE. See the GNU Lesser General Public License for more details.
%
% You should have received a copy of the GNU Lesser General Public License along with the Attempto
% Parsing Engine (APE). If not, see http://www.gnu.org/licenses/.


/* APE client for the Unix commandline.

Building the executable file (stack sizes are in kilobytes):

==
swipl -g "working_directory(_, 'parser'), [fit_to_plp], halt."
swipl -O -F none -g "[ape], qsave_program('ape.exe', [goal(ape), toplevel(halt), local(25000), global(50000)])." -t halt
==

Note that you can use smaller stack sizes if needed. It is known that global stack of 25600 was not
enough to parse 1687 sentences of the Hydrology verbalization.

Command-line arguments of the executable:
==
-text "TEXT"     The input ACE text. If neither -text nor -file is present then the ACE text is read from stdin.
-file FILENAME   The input file containing the ACE text.
-ulex FILENAME   The user lexicon file to be loaded.
-solo OUTPUT     Output just one output component. OUTPUT has to be one of {paraphrase,paraphrase1,paraphrase2,owlfss,
                 owlrdf,owlxml,fol,pnf,tokens,syntax,drs,drsxml,drspp,drshtml,syntaxpp}.
-cdrs            Output the DRS as a Prolog term.
-cdrsxml         Output the DRS in XML.
-cdrspp          Output the DRS in pretty-printed form in plain text.
-cdrshtml        Output the DRS in pretty-printed form in HTML.
-cparaphrase     Output a paraphrase which is a "best-effort" combination of paraphrase1 and paraphrase2.
-cparaphrase1    Output a paraphrase which uses full sentences instead of relative clauses.
-cparaphrase2    Output a paraphrase which uses relative clauses instead of full sentences.
-ctokens         Output tokens as a Prolog list of lists.
-csentences      Output sentences as a Prolog list.
-csyntax         Output syntax trees as a Prolog list.
-csyntaxpp       Output syntax trees in pretty-printed form.
-cowlfss         Output OWL 2 in the Functional-Style Syntax representation.
-cowlrdf         Output OWL 2 in the RDF/XML representation.
-cowlxml         Output OWL 2 in the XML representation.
-cfol            Output standard first-order logic representations (default form and prenex normal form) of the DRS as
                 a Prolog term.
-uri URI         URI for the OWL outputs.
-guess           Guess the word-class of unknown words.
-help            Shows this help page.
==

*/

:- use_module(get_ape_results).

:- set_prolog_flag(encoding, utf8).

:- initialization on_signal(int, _, default).


%% argument(?Arg, -Value, -Desc)
%
% @tbd maybe we could describe the types more formally, to that they could be checked,
% e.g. the type of the solo value is one of [drs, drsxml, ...], which could be checked
% with memberchk/2.
%

argument('-text', '"TEXT"', 'The input ACE text. If neither -text nor -file is present then the ACE text is read from stdin.').
argument('-file', 'FILENAME', 'The input file containing the ACE text.').
argument('-ulex', 'FILENAME', 'The user lexicon file to be loaded.').
argument('-solo', 'OUTPUT', 'Output just one output component. OUTPUT has to be one of {paraphrase,paraphrase1,paraphrase2,owlfss,owlrdf,owlxml,fol,pnf,tokens,syntax,drs,drsxml,drspp,drshtml,syntaxpp}.').
argument('-cinput', '', hidden).
argument('-cdrs', '', 'Output the DRS as a Prolog term.').
argument('-cdrsxml', '', 'Output the DRS in XML.').
argument('-cdrspp', '', 'Output the DRS in pretty-printed form in plain text.').
argument('-cdrshtml', '', 'Output the DRS in pretty-printed form in HTML.').
argument('-cparaphrase', '', 'Output a paraphrase which is a "best-effort" combination of paraphrase1 and paraphrase2.').
argument('-cparaphrase1', '', 'Output a paraphrase which uses full sentences instead of relative clauses.').
argument('-cparaphrase2', '', 'Output a paraphrase which uses relative clauses instead of full sentences.').
argument('-ctokens', '', 'Output tokens as a Prolog list of lists.').
argument('-csentences', '', 'Output sentences as a Prolog list.').
argument('-csyntax', '', 'Output syntax trees as a Prolog list.').
argument('-csyntaxpp', '', 'Output syntax trees in pretty-printed form.').
argument('-cowlfss', '', 'Output OWL 2 in the Functional-Style Syntax representation.').
argument('-cowlrdf', '', 'Output OWL 2 in the RDF/XML representation.').
argument('-cowlxml', '', 'Output OWL 2 in the XML representation.').
argument('-cfol', '', 'Output standard first-order logic representations (default form and prenex normal form) of the DRS as a Prolog term.').
argument('-uri', 'URI', 'URI for the OWL outputs.').
argument('-guess', '', 'Guess the word-class of unknown words.').
argument('-help', '', 'Shows this help page.').


%% ape

ape :-
	current_prolog_flag(argv, RawArgList),
	get_arglist(RawArgList, ArgList),
	arglist_namevaluelist(ArgList, InputList1),
	process_input(InputList1).


%% get_arglist(+RawArgList, -ArgList)

get_arglist(RawArgList, ArgList) :-
    append(_, ['--'|ArgList], RawArgList),
    !.

get_arglist([_|ArgList], ArgList).


%% process_input(+InputList)

process_input(InputList) :-
    ( InputList = [] ; member(help=on, InputList) ),
    !,
    show_help.

process_input(InputList1) :-
	read_file(InputList1, InputList2),
	get_ape_results(InputList2, Content),
	write(Content), nl.


%% show_help

show_help :-
    write('Attempto Parsing Engine for ACE 6.0\n'),
    write('Copyright 2008 Attempto Group, University of Zurich\n'),
    write('This program comes with ABSOLUTELY NO WARRANTY.\n'),
    write('This is free software, and you are welcome to redistribute it under certain conditions.\n'),
    write('Please visit http://attempto.ifi.uzh.ch for details.\n'),
    nl,
    write('Command-line arguments:\n'),
    argument(Arg, Value, Desc),
    \+ Desc = hidden,
    format('~w ~w~17|~w~n', [Arg, Value, Desc]),
    fail.

show_help.


%% arglist_namevaluelist(+ArgList:list, -NameValueList:list) is det.
%
% @param ArgList is a list of arguments
% @param NameValueList is a list of ArgumentName=ArgumentValue pairs

arglist_namevaluelist([], []).

arglist_namevaluelist([Arg|Tail1], [Name=on|Tail2]) :-
    argument(Arg, '', _),
    !,
    atom_concat('-', Name, Arg),
	arglist_namevaluelist(Tail1, Tail2).

arglist_namevaluelist([Arg,ValueAtom|Tail1], [Name=Value|Tail2]) :-
    argument(Arg, _, _),
    \+ argument(ValueAtom, _, _),
    !,
    atom_concat('-', Name, Arg),
	(
		catch(atom_number(ValueAtom, ValueNumber), _, fail)
	->
		Value = ValueNumber
	;
		Value = ValueAtom
	),
	arglist_namevaluelist(Tail1, Tail2).

arglist_namevaluelist([Arg|_], _) :-
    argument(Arg, _, _),
    !,
    format(user_error, 'ERROR: Missing value for argument: ~w~n', Arg),
    fail.

arglist_namevaluelist([Arg|_], _) :-
    format(user_error, 'ERROR: Illegal argument: ~w~n', Arg),
    fail.


%% read_file(+InputListIn, -InputListOut)

read_file(InputList, [text=AceText | InputList]) :-
    member(file=AceFile, InputList),
    !,
    read_file_to_codes(AceFile, AceTextCodes, [encoding(utf8)]),
    atom_codes(AceText, AceTextCodes).

read_file(InputList, InputList) :-
    member(text=_, InputList),
    !.

read_file(InputList, [text=AceText | InputList]) :-
    prompt(_, ''),
    read_stream_to_codes(user_input, AceTextCodes),
    atom_codes(AceText, AceTextCodes).
