% This file is part of the Attempto Parsing Engine (APE).
% Copyright 2008, Attempto Group, University of Zurich (see http://attempto.ifi.uzh.ch).
%
% The Attempto Parsing Engine (APE) is free software: you can redistribute it and/or modify it
% under the terms of the GNU Lesser General Public License as published by the Free Software
% Foundation, either version 3 of the License, or (at your option) any later version.
%
% The Attempto Parsing Engine (APE) is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
% PURPOSE. See the GNU Lesser General Public License for more details.
%
% You should have received a copy of the GNU Lesser General Public License along with the Attempto
% Parsing Engine (APE). If not, see http://www.gnu.org/licenses/.


:- module(get_ape_results, [
		get_ape_results_timelimit/3,
		get_ape_results_timelimit/4,
		get_ape_results/2,
		get_ape_results/3,
		load_ulex/1,
		get_content/4
	]).


/** <module> Interface for the ACE tools (ACE parser, DRS verbalizer, ...)

@author Kaarel Kaljurand
@author Tobias Kuhn
@version 2008-04-09

Usage with multiple results returned (i.e. multi-mode):

==
get_ape_results([text='Every man waits.', cparaphrase1=on], ContentType, Content).
get_ape_results([text='Every man waits.', cparaphrase=on, cparaphrase1=on], ContentType, Content).
get_ape_results([text='A man waits.', cinput=on, cdrs=on, cdrspp=on, cparaphrase=on, cparaphrase1=on, cparaphrase2=on, ctokens=on, csyntax=on, csyntaxpp=on, cfol=on], ContentType, Content).
==

Usage with a single result returned (i.e. solo-mode):

==
get_ape_results([text='A man sees a dog.', solo=drs], ContentType, Content).
get_ape_results([text='Every man owns a dog.', solo=owlrdf], ContentType, Content).
get_ape_results([text='Every man owns a dog.', solo=owlfss], ContentType, Content).
get_ape_results([text='Peeter likes Mary.', solo=owlrdf], ContentType, Content).
get_ape_results([text='Peeter likes Mary.', solo=owlfss], ContentType, Content).
get_ape_results([text='Every man waits.', solo=owlrdf], ContentType, Content).
get_ape_results([text='Every man waits.', solo=owlfss], ContentType, Content).
==

Note that returning of the typed DRS can be turned on by adding "typed=on" to the query.

@tbd: share the calls between solo and multi
@tbd: provide all outputs
	(1) as serialized Prolog terms,
	(2) as pretty-printed terms,
	(3) as XML, JSON, HTML, etc.
*/

% Default encoding used for opening files in text mode.
:- set_prolog_flag(encoding, utf8).

% Note: The following line fixes the testcase: 3.1415926536 approximates Pi.
% Note that using 'f' instead of 'g' does not drop the trailing zeros.
% @bug: sometimes weird digits are attached to the end.
:- set_prolog_flag(float_format, '%.11g').

% Richard A. O'Keefe: to see full precision for IEEE doubles, do
% :- set_prolog_flag(float_format, '%.18g').
% So it seems that values above 18 do not make sense.

:- assert(user:file_search_path(ape, '.')).


:- use_module(ape('utils/morphgen'), [
		acesentencelist_pp/2
	]).

:- use_module(ape('utils/ace_niceace'), [
		tokens_to_sentences/2
	]).

:- use_module(ape('utils/typeddrs_to_untypeddrs')).
:- use_module(ape('utils/drs_to_xml')).
:- use_module(ape('utils/drs_to_fol_to_prenex')).
:- use_module(ape('utils/drs_to_ascii')).
:- use_module(ape('utils/drs_to_ace')).
:- use_module(ape('utils/drs_to_coreace')).
:- use_module(ape('utils/drs_to_npace')).
:- use_module(ape('utils/drs_to_html')).
:- use_module(ape('utils/drs_to_ruleml')).
:- use_module(ape('utils/trees_to_ascii')).
:- use_module(ape('utils/owlswrl/drs_to_owlswrl')).
:- use_module(ape('utils/owlswrl/owlfss_owlrdfxml')).
:- use_module(ape('utils/owlswrl/owlfss_owlxml')).
:- use_module(ape('utils/owlswrl/owlfss_ascii')).
:- use_module(ape('lexicon/dynamic_ulex')).
:- use_module(ape('parser/ace_to_drs')).
:- use_module(ape('logger/error_logger')).

:- use_module(ape('utils/xmlterm_to_xmlatom'), [
		xmlterm_to_xmlatom/2,
		xmlterm_to_xmlatom/3
	]).

:- use_module(ape('utils/serialize_term'), [
		serialize_term_into_atom/2
	]).


:- style_check(-singleton).
:- style_check(-discontiguous).
:- use_module(ape('lexicon/clex')).
:- style_check(+discontiguous).
:- style_check(+singleton).


%% get_form_value(+Form:list, +Key:atom, -Value:atom) is det.
%
% Value is the value of the Key in the Form or '' if Key is not found.
%
% @param Form is a list of pairs in the form of Key=Value
% @param Key is a key
% @param Value is a value
%
get_form_value(Form, Key, Value) :-
	memberchk(Key=Value, Form),
	!.

get_form_value(_Form, _Key, '').


%% get_ape_results_timelimit(+Input:list, -Content:atom, +TimeLimit) is det.
%% get_ape_results_timelimit(+Input:list, -ContentType:atom, -Content:atom, +TimeLimit) is det.
%
% There is call_with_time_limit(+Time, :Goal) defined in library(time),
% part of clib package. On timeout this throws the exception time_limit_exceeded.
% But we catch other exceptions as well...
%
% @param Input is a list of input parameters (e.g. an ACE text, a user lexicon name)
% @param ContentType is one of {text/plain, text/xml}
% @param Content is the returned XML
% @param TimeLimit the timelimit in seconds
%
get_ape_results_timelimit(Input, Content, TimeLimit) :-
	get_ape_results_timelimit(Input, _ContentType, Content, TimeLimit).

get_ape_results_timelimit(Input, ContentType, Content, TimeLimit) :-
	catch(
		call_with_time_limit(
			TimeLimit,
			get_ape_results(Input, ContentType, Content)
		),
		CatchType,
		catchtype_errormessage(CatchType, ContentType, Content)
	).


%% catchtype_errormessage(+CatchType:atom, -ContentType:atom, -ErrorMessage:atom) is det.
%
% Returns an error message as XML. The error message is set by catch/3,
% this predicate just formats the message.
% This is a very toplevel error message indicating either that the time limit
% was exceeded or that there were resource errors (out of stack, etc.) or programmer
% errors (undefined predicate called, etc.).
%
% @param CatchType is one of {time_limit_exceeded, ...}
% @param ContentType is always 'text/xml'
% @param ErrorMessage is an end-user-level message that corresponds to the CatchType
%
catchtype_errormessage(
	time_limit_exceeded,
	'text/xml',
	'<apeResult><messages><message importance="error" type="ws" sentence="" token="" value="time_limit_exceeded"
repair="Split the text into smaller parts and parse them separately."/></messages></apeResult>'
	) :- !.

catchtype_errormessage(CatchType, 'text/xml', ErrorMessage) :-
	format(atom(CatchTypeAsAtom), "~w", [CatchType]),
	xmlterm_to_xmlatom(element(apeResult, [], [
			element(messages, [], [
				element(message, [
					importance='error',
					type='ws',
					sentence='',
					token='',
					value=CatchTypeAsAtom,
					repair='Fatal error. Please send screenshot to APE developers.'
					], [])
			])
		]), ErrorMessage).


%% get_ape_results(+Input:list, -Content:atom) is det.
%% get_ape_results(+Input:list, -ContentType:atom, -Content:atom) is det.
%
% @param Input is a list of input parameters (e.g. an ACE text, a user lexicon name)
% @param ContentType is one of {text/plain, text/xml}
% @param Content is the returned XML
%
get_ape_results(Input, Content) :-
	get_ape_results(Input, _ContentType, Content).

get_ape_results(Input, ContentType, Content) :-
	% just to be sure:
	discard_ulex,

	clear_messages,

	load_ulex(Input),

	get_form_value(Input, text, ACEText),
	%%%get_form_value(Input, clex, ClexOnOff),
	get_form_value(Input, guess, GuessOnOff),

	ace_to_drs:ace_to_drs_time(
		ACEText, [clex:on, guess:GuessOnOff],
		Tokens, Syntaxtrees, Drs, Messages, [DT, DP, DR]
	),

	(
		get_form_value(Input, uri, UriTmp),
		UriTmp \= ''
	->
		Uri = UriTmp
	;
		Uri = 'http://attempto.ifi.uzh.ch/ontologies/owlswrl/test'
	),

	get_content(Input, [
		[DT, DP, DR],
		ACEText,
		Tokens,
		Drs,
		Syntaxtrees,
		Messages,
		Uri
	], ContentType, Content),
	
	discard_ulex,
	!.


% @bug: This is sometimes called, but we should know why.
% It must be here, just to catch errors.
get_ape_results(_, 'text/plain', '') :-
	discard_ulex.


%% get_content(+Form, +Input:list, -ContentType:atom, -Content:atom) is det.
%
%
get_content(Form, Input, ContentType, Content) :-
	get_form_value(Form, solo, SoloType),
	SoloType \= '',
	!,
	get_solo_content(SoloType, Form, Input, ContentType, Content).

get_content(Form, Input, ContentType, Content) :-
	get_multi_content(Form, Input, ContentType, Content).


%% get_solo_content(+SoloType:atom, +Input:list, -ContentType:atom, -Content:atom) is det.
%
% @param SoloType is one of {syntax, syntaxpp, drs, drspp, drsxml, drshtml,
%	owlrdf, owlxml, owlfss,	ruleml, fol, pnf,
%	paraphrase, paraphrase1, paraphrase2, tokens}
% @param Input is a list of APE outputs
% @param ContentType is one of {text/xml, text/plain}
% @param Content is the result (e.g. a syntax tree or a paraphrase) in XML or plain text
%
% @deprecated parameter: input

% If there are APE error messages
% then we print the messages and not the solo output.
% Note: we do not care about the warning messages.
get_solo_content(_, _, [_, _, _, _, _, _, _], 'text/xml', Content) :-
	get_error_messages([M | ErrorMessages]),
	!,
	messages_xmlmessages([M | ErrorMessages], XmlErrorMessages),
	xmlterm_to_xmlatom(element(messages, [], XmlErrorMessages), Content).

% @deprecated
% @bug commented out nobody uses it (note that multi input is used by APE web client
% for the regression test database)
/*
get_solo_content(input, _, [_, AceText | _], 'text/plain', Content) :-
	serialize_term_into_atom(AceText, Content).
*/

get_solo_content(syntax, _, [_, _, _, _, Syntax | _], 'text/plain', Content) :-
	serialize_term_into_atom(Syntax, Content).

get_solo_content(syntaxpp, _, [_, _, _, _, Syntax | _], 'text/plain', AsciiTrees) :-
	trees_to_ascii:trees_to_ascii(Syntax, AsciiTrees).

get_solo_content(drs, Form, [_, _, _, Drs | _], 'text/plain', Content) :-
	untype_if_needed(Form, Drs, Drs2),
	serialize_term_into_atom(Drs2, Content).

get_solo_content(drsxml, Form, [_, _, _, Drs | _], 'text/xml', DrsXml) :-
	untype_if_needed(Form, Drs, Drs2),
	drs_to_xmlatom(Drs2, DrsXml).

get_solo_content(drspp, Form, [_, _, _, Drs | _], 'text/plain', DrsAscii) :-
	untype_if_needed(Form, Drs, Drs2),
	drs_to_ascii:drs_to_ascii(Drs2, DrsAscii).

get_solo_content(drshtml, Form, [_, _, _, Drs | _], 'text/xml', DrsHtml) :-
	untype_if_needed(Form, Drs, Drs2),
	drs_to_html:drs_to_html(Drs2, DrsHtml).

% @bug: we are using ignore/1 here, but cleaner would be to make sure that ACE->OWL never fails.
get_solo_content(owlrdf, _, [_, AceText, _, Drs, _, _, Uri], 'text/xml', Content) :-
	typeddrs_to_untypeddrs:typeddrs_to_untypeddrs(Drs, UntypedDrs),
	ignore(drs_to_owlswrl:drs_to_owlswrl(UntypedDrs, Uri, AceText, OwlFss)),
	get_messages_with_type(owl, Messages),
	(
		Messages = []
	->
		owlfss_owlrdfxml:owlfss_owlrdfxml(OwlFss, OwlRdfxml),
		xmlterm_to_xmlatom(OwlRdfxml, Content)
	;
		messages_xmlmessages(Messages, XmlMessages),
		xmlterm_to_xmlatom(element(messages, [], XmlMessages), Content)
	).

% @bug: we are using ignore/1 here, but cleaner would be to make sure that ACE->OWL never fails.
get_solo_content(owlxml, _, [_, AceText, _, Drs, _, _, Uri], 'text/xml', Content) :-
	typeddrs_to_untypeddrs:typeddrs_to_untypeddrs(Drs, UntypedDrs),
	ignore(drs_to_owlswrl:drs_to_owlswrl(UntypedDrs, Uri, AceText, OwlFss)),
	get_messages_with_type(owl, Messages),
	(
		Messages = []
	->
		(
			(OwlFss = 'Ontology'(_, Axioms), member('Implies'(_, _), Axioms))
		->
			owlfss_owlrdfxml:owlfss_owlrdfxml(OwlFss, OwlXml)
		;
			owlfss_owlxml:owlfss_owlxml(OwlFss, OwlXml)
		),
		xmlterm_to_xmlatom(OwlXml, Content)
	;
		messages_xmlmessages(Messages, XmlMessages),
		xmlterm_to_xmlatom(element(messages, [], XmlMessages), Content)
	).

% @bug: we are using ignore/1 here, but cleaner would be to make sure that ACE->OWL never fails.
get_solo_content(owlfss, _, [_, AceText, _, Drs, _, _, Uri], MimeType, Content) :-
	typeddrs_to_untypeddrs:typeddrs_to_untypeddrs(Drs, UntypedDrs),
	ignore(drs_to_owlswrl:drs_to_owlswrl(UntypedDrs, Uri, AceText, OwlFss)),
	get_messages_with_type(owl, Messages),
	(
		Messages = []
	->
		serialize_term_into_atom(OwlFss, Content),
		MimeType = 'text/plain'
	;
		messages_xmlmessages(Messages, XmlMessages),
		xmlterm_to_xmlatom(element(messages, [], XmlMessages), Content),
		MimeType = 'text/xml'
	).

get_solo_content(ruleml, _, [_, _, _, Drs | _], 'text/xml', RulemlDrs) :-
	print_ruleml(Drs, [element(_, [], [RulemlDrs])]).

get_solo_content(fol, Form, [_, _, _, Drs | _], 'text/plain', Content) :-
	untype_if_needed(Form, Drs, Drs2),
	drs_fol_pnf:drs_fol(Drs2, Fol),
	!,
	serialize_term_into_atom(Fol, Content).

get_solo_content(pnf, Form, [_, _, _, Drs | _], 'text/plain', Content) :-
	untype_if_needed(Form, Drs, Drs2),
	drs_fol_pnf:drs_pnf(Drs2, Pnf),
	!,
	serialize_term_into_atom(Pnf, Content).

get_solo_content(paraphrase, _, [_, _, _, Drs | _], 'text/plain', AceText) :-
	typeddrs_to_untypeddrs:typeddrs_to_untypeddrs(Drs, UntypedDrs),
	drs_to_ace:drs_to_ace(UntypedDrs, AceSentenceList),
	get_acetext_or_error(AceSentenceList, AceText).

get_solo_content(paraphrase1, _, [_, _, _, Drs | _], 'text/plain', AceText) :-
	typeddrs_to_untypeddrs:typeddrs_to_untypeddrs(Drs, UntypedDrs),
	drs_to_coreace:drs_to_coreace(UntypedDrs, AceSentenceList),
	get_acetext_or_error(AceSentenceList, AceText).

get_solo_content(paraphrase2, _, [_, _, _, Drs | _], 'text/plain', AceText) :-
	typeddrs_to_untypeddrs:typeddrs_to_untypeddrs(Drs, UntypedDrs),
	drs_to_npace:drs_to_npace(UntypedDrs, AceSentenceList),
	get_acetext_or_error(AceSentenceList, AceText).

get_solo_content(tokens, _, [_, _, TokenList | _], 'text/plain', Content) :-
	serialize_term_into_atom(TokenList, Content).


% @bug: this is reached with the URL:
% http://attempto.ifi.uzh.ch/ws/ape/apews.perl?text=Where+does+a+man+wait%3F&solo=owlrdf
% (Not anymore. But the error message is strange. Note that the period is missing
% in the sentence, but the error message doesn't discuss this.)
get_solo_content(_, _, _, 'text/plain', 'ERROR: Wrong solo type or maybe an APE bug.').


%% get_multi_content(+Input:list, -ContentType:atom, -Content:atom) is det.
%
% @bug: we ignore the parser+refres messages, and get a fresh list
% of messages from the messages repository (it will include the parser+refres
% messages anyway).
%
get_multi_content(Form, [
		[DT, DP, DR],
		ACEText,
		Tokens,
		Drs,
		Syntaxtrees,
		_Messages,
		Uri
	], 'text/xml', Content) :-

	print_duration(DT, DP, DR, XmlDuration),

	typeddrs_to_untypeddrs:typeddrs_to_untypeddrs(Drs, UntypedDrs),

	(
		get_form_value(Form, typed, on)
	->
		PossiblyUntypedDrs = Drs
	;
		PossiblyUntypedDrs = UntypedDrs
	),


	multi_caller(Form, cinput, print_input_text(ACEText, XmlInputText), XmlInputText),
	multi_caller(Form, ctokens, print_tokens(Tokens, XmlTokens), XmlTokens),
	multi_caller(Form, csentences, print_sentences(Tokens, XmlSentences), XmlSentences),

	multi_caller(Form, cdrs, print_drs(PossiblyUntypedDrs, drs, XmlDrs), XmlDrs),

	multi_caller(Form, cdrsrt, print_drs_rt(Drs, drsrt, XmlDrsrt), XmlDrsrt),

	multi_caller(Form, csyntax, print_syntax_list(Syntaxtrees, XmlSyntax), XmlSyntax),
	multi_caller(Form, cparaphrase, print_paraphrase(UntypedDrs, XmlParaphrase), XmlParaphrase),
	multi_caller(Form, cparaphrase1, print_paraphrase1(UntypedDrs, XmlParaphrase1), XmlParaphrase1),
	multi_caller(Form, cparaphrase2, print_paraphrase2(UntypedDrs, XmlParaphrase2), XmlParaphrase2),

	multi_caller(Form, cdrspp, print_drs_pp(PossiblyUntypedDrs, drspp, XmlDrsPp), XmlDrsPp),

	multi_caller(Form, cdrsxml, print_drs_xml(PossiblyUntypedDrs, drsxml, XmlDrsXml), XmlDrsXml),

	multi_caller(Form, cdrshtml, print_drs_html(PossiblyUntypedDrs, drshtml, XmlDrsHtml), XmlDrsHtml),

	multi_caller(Form, csyntaxpp, print_syntaxpp(Syntaxtrees, XmlSyntaxPp), XmlSyntaxPp),

	multi_caller(Form, cowlrdf, print_owlrdf(UntypedDrs, Uri, XmlOwlRdf), XmlOwlRdf),
	multi_caller(Form, cowlfss, print_owlfss(UntypedDrs, Uri, XmlOwlFss), XmlOwlFss),
	multi_caller(Form, cowlxml, print_owlxml(UntypedDrs, Uri, XmlOwlXml), XmlOwlXml),

	multi_caller(Form, cruleml, print_ruleml(Drs, XmlRuleml), XmlRuleml),

	multi_caller(Form, cfol, print_fol(PossiblyUntypedDrs, XmlFol), XmlFol),

	print_messages(XmlMessages),

	flatten([XmlDuration, XmlInputText, XmlTokens, XmlSentences,
		XmlDrs, XmlDrsrt, XmlSyntax, XmlDrsPp, XmlDrsXml, XmlDrsHtml,
		XmlSyntaxPp, XmlParaphrase, XmlParaphrase1, XmlParaphrase2,
		XmlOwlRdf, XmlOwlFss, XmlOwlXml, XmlRuleml, XmlFol, XmlMessages], XmlOutput),
	xmlterm_to_xmlatom(element(apeResult, [], XmlOutput), [header(true)], Content).


%% multi_caller(+Form:list, +InputType:atom, :Goal:term, -Result:term) is det.
%
% @param Form is a list of input parameters
% @param InputType is an input type, one of {drs, syntaxpp, ...}
% @param Goal is a goal to be called
% @param Result is a result to be delivered
%
multi_caller(Form, InputType, Goal, Result) :-
	(
		get_form_value(Form, InputType, on)
	->
		call(Goal)
	;
		Result = []
	),
	!.

multi_caller(_Form, _InputType, _Goal, []).


%% print_duration(+DTokenizer:number, +DParser:number, +DRefres:number, -XmlDuration:xml) is det.
%
% Returns the duration of tokenization, parsing and reference resolution
% wrapped in XML.
%
print_duration(DTokenizer, DParser, DRefres, [element(duration, [tokenizer=T, parser=P, refres=R], [])]) :-
	with_output_to(atom(T), format("~3f", [DTokenizer])),
	with_output_to(atom(P), format("~3f", [DParser])),
	with_output_to(atom(R), format("~3f", [DRefres])).


%% print_input_text(+ACEText:atom, -XmlAceText:xml) is det.
%
% Returns the ACE text wrapped in XML.
%
% @deprecated
%
print_input_text(ACEText, [element(acetext, [], [ACEText])]).


%% print_tokens(+Tokens:list, -XmlTokens:xml) is det.
%
% Returns the tokens wrapped in XML.
%
print_tokens(Tokens, [element(tokens, [], [Content])]) :-
	serialize_term_into_atom(Tokens, Content).


%% print_sentences(+Tokens:list, -XmlSentences:xml) is det.
%
% Converts the list of sentences (each a list of tokens) into
% a list of sentences (each an atom).
% Returns the result wrapped in XML.
%
print_sentences(Tokens, [element(sentences, [], [Content])]) :-
	tokens_to_sentences(Tokens, Sentences),
	serialize_term_into_atom(Sentences, Content).


%% print_syntax_list(+Syntaxtrees:list, -XmlTrees:xml) is det.
%
% Returns the syntax trees wrapped in XML.
%
% @bug: used to be: sformat(XmlTrees, '<syntax><![CDATA[~q]]></syntax>', [Syntaxtrees]).
%
print_syntax_list(Syntaxtrees, [element(syntax, [], [Content])]) :-
	serialize_term_into_atom(Syntaxtrees, Content).


%% print_fol(+Drs:term, -XmlFol:xml) is det.
%
% Returns FOL and PNF wrapped in XML.
%
% @param Drs is an Attempto DRS
%
print_fol(Drs, [element(fol, [], [FolContent]), element(pnf, [], [PnfContent])]) :-
	drs_fol_pnf:drs_fol_pnf(Drs, Fol, Pnf),
	drs_fol_pnf:fol_to_folpp(Fol, FolContent),
	drs_fol_pnf:fol_to_folpp(Pnf, PnfContent).


%% print_paraphrase(+Drs:term, -XmlParaphrase:xml) is det.
%
% Returns the paraphrase wrapped in XML.
%
% This paraphrase is a the same as paraphrase2 unless generating it
% fails, in this case paraphrase1 is used.
%
% @param Drs is an Attempto DRS
%
print_paraphrase(Drs, [element(paraphrase, [], [AceText])]) :-
	Drs = drs(_, [_ | _]),
	drs_to_ace:drs_to_ace(Drs, AceSentenceList),
	get_acetext_or_error(AceSentenceList, AceText).


%% print_paraphrase1(+Drs:term, -XmlParaphrase:xml) is det.
%
% Returns the paraphrase1 wrapped in XML.
%
% @param Drs is an Attempto DRS
%
print_paraphrase1(Drs, [element(paraphrase1, [], [AceText])]) :-
	Drs = drs(_, [_ | _]),
	drs_to_coreace:drs_to_coreace(Drs, AceSentenceList),
	get_acetext_or_error(AceSentenceList, AceText).


%% print_paraphrase2(+Drs:term, -XmlParaphrase:xml) is det.
%
% Returns the paraphrase2 wrapped in XML.
%
% @param Drs is an Attempto DRS
%
print_paraphrase2(Drs, [element(paraphrase2, [], [AceText])]) :-
	Drs = drs(_, [_ | _]),
	drs_to_npace:drs_to_npace(Drs, AceSentenceList),
	get_acetext_or_error(AceSentenceList, AceText).


%% print_owlrdf(+Drs:term, +Uri:atom, -OwlRdfXml:xml) is det.
%
% Returns OWL/SWRL in RDF/XML Syntax wrapped in XML.
%
% @bug: I'm not sure we have to convert the XML into an atom (because
% it is later done anyway). But it doesn't work otherwise...
%
% @param Drs is an Attempto DRS
%
print_owlrdf(Drs, Uri, [element(owlrdf, [], [Content])]) :-
	drs_to_owlswrl:drs_to_owlswrl(Drs, Uri, OwlFss),
	owlfss_owlrdfxml:owlfss_owlrdfxml(OwlFss, OwlRdfxml),
	xmlterm_to_xmlatom(OwlRdfxml, Content).


%% print_owlfss(+Drs:term, +Uri:atom, -OwlFss:xml) is det.
%
% Returns OWL/SWRL in Functional-Style Syntax pretty-printed
% and wrapped in XML.
%
% @param Drs is an Attempto DRS
%
print_owlfss(Drs, Uri, [element(owlfss, [], [Content])]) :-
	drs_to_owlswrl:drs_to_owlswrl(Drs, Uri, OwlFss),
	owlfss_ascii:owlfss_ascii(OwlFss, Content).


%% print_owlxml(+Drs:term, +Uri:atom, -OwlXml:xml) is det.
%
% Returns OWL in XML Syntax and SWRL in RDF/XML wrapped in XML.
%
% @param Drs is an Attempto DRS
%
print_owlxml(Drs, Uri, [element(owlxml, [], [Content])]) :-
	drs_to_owlswrl:drs_to_owlswrl(Drs, Uri, OwlFss),
	(
		(OwlFss = 'Ontology'(_, Axioms), member('Implies'(_, _), Axioms))
	->
		owlfss_owlrdfxml:owlfss_owlrdfxml(OwlFss, OwlXml)
	;
		owlfss_owlxml:owlfss_owlxml(OwlFss, OwlXml)
	),
	xmlterm_to_xmlatom(OwlXml, Content).


%% print_ruleml(+Drs:term, -Ruleml:xml) is det.
%
% Returns the RuleML/folog wrapped in XML.
%
% @param Drs is an Attempto DRS
%
print_ruleml(Drs, [element(ruleml, [], [Content])]) :-
	drs_to_ruleml:drs_to_ruleml(Drs, Ruleml),
	xmlterm_to_xmlatom([Ruleml], Content).


%% print_drs(+Drs:term, +TagName:atom, -XmlDrs:xml) is det.
%
% Returns the DRS serialized and wrapped in XML.
%
% @param Drs is an Attempto DRS
%
print_drs(Drs, TagName, [element(TagName, [], [Content])]) :-
	serialize_term_into_atom(Drs, Content).


%% print_drs_rt(+Drs:term, +TagName:atom, -XmlDrs:xml) is det.
%
% Returns the "DRS for regression tester" wrapped in XML.
%
% @param Drs is an Attempto DRS
%
% @deprecated: use print_drs/3
%
print_drs_rt(Drs, TagName, [element(TagName, [], [Content])]) :-
	print_drs(Drs, TagName, [element(TagName, [], [Content])]).


%% print_drs_pp(+Drs:term, +TagName:atom, -XmlDrs:xml) is det.
%
% Returns the pretty-printed DRS wrapped in XML.
%
% @param Drs is an Attempto DRS
%
print_drs_pp(Drs, TagName, [element(TagName, [], [Content])]) :-
	drs_to_ascii:drs_to_ascii(Drs, Content).


%% print_drs_xml(+Drs:term, +TagName:atom, -XmlDrs:xml) is det.
%
% Returns the DRS marked-up in XML.
%
% @param Drs is an Attempto DRS
%
print_drs_xml(Drs, TagName, [element(TagName, [], [DrsXml])]) :-
	drs_to_xmlatom(Drs, DrsXml).


%% print_drs_html(+Drs:term, +TagName:atom, -HtmlDrs:xml) is det.
%
% Returns the DRS pretty-printed in HTML.
%
% @param Drs is an Attempto DRS
%
print_drs_html(Drs, TagName, [element(TagName, [], [Content])]) :-
	drs_to_html:drs_to_html(Drs, Content).


%% print_syntaxpp(+Syntaxtrees:list, -XmlTrees:xml) is det.
%
% Returns the pretty-printed syntax trees wrapped in XML.
%
print_syntaxpp(Syntaxtrees, [element(syntaxpp, [], [AsciiTrees])]) :-
	trees_to_ascii:trees_to_ascii(Syntaxtrees, AsciiTrees).


%% print_messages(-XmlMessages:xml) is det.
%
% @param XmlMessages are APE warning/error messages in XML (SWI-Prolog notation)
%
print_messages(element(messages, [], XmlMessages)) :-
	get_messages_in_xml(XmlMessages).


%% load_ulex(+FormInput:list) is det.
%
%
load_ulex(FormInput) :-
	get_form_value(FormInput, ulex, AceLexicon),
	AceLexicon = '',
	!.

load_ulex(FormInput) :-
	get_form_value(FormInput, ulex, AceLexicon),
	read_ulex(AceLexicon).


%% untype_if_needed(+Form:list, +TypedDrs:term, -UntypedDrs:term) is det.
%
%
untype_if_needed(Form, TypedDrs, TypedDrs) :-
	get_form_value(Form, typed, on),
	!.

untype_if_needed(_Form, TypedDrs, UntypedDrs) :-
	typeddrs_to_untypeddrs:typeddrs_to_untypeddrs(TypedDrs, UntypedDrs).


%% get_acetext_or_error(+AceSentenceList:list, -AceTextOrError:atom) is det.
%
% If AceSentenceList is an empty list then returns an error message.
% Otherwise returns a pretty-printed ACE text.
%
% @param AceSentenceList is a list of ACE sentences
% @param AceTextOrError is an ACE text or an error message
%
get_acetext_or_error([], 'NOT IMPLEMENTED') :-
	!.

get_acetext_or_error(AceSentenceList, AceText) :-
	acesentencelist_pp(AceSentenceList, AceText).
