% This file is part of the Attempto Parsing Engine (APE).
% Copyright 2008, Attempto Group, University of Zurich (see http://attempto.ifi.uzh.ch).
%
% The Attempto Parsing Engine (APE) is free software: you can redistribute it and/or modify it
% under the terms of the GNU Lesser General Public License as published by the Free Software
% Foundation, either version 3 of the License, or (at your option) any later version.
%
% The Attempto Parsing Engine (APE) is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
% PURPOSE. See the GNU Lesser General Public License for more details.
%
% You should have received a copy of the GNU Lesser General Public License along with the Attempto
% Parsing Engine (APE). If not, see http://www.gnu.org/licenses/.


:- module(acelexicon_apelex, [
		acelexicon_apelex/2  % +AceLexiconEntry, -ApeLexEntries
	]).

:- use_module(chars).


/** <module> ACE Lexicon to Apelex Converter

This module converts from the ACE Lexicon format into the internal Apelex format.

@author Tobias Kuhn
@version 2008-03-13
*/


%% acelexicon_apelex(+AceLexiconEntry, -ApeLexEntries)
%
% Converts an ACE Lexicon entry into a list of Apelex entries.

acelexicon_apelex(AceLexiconEntry, ApeLexEntries) :-
    findall(ApeLexEntry, convert(AceLexiconEntry, ApeLexEntry), ApeLexEntries). 


%% convert(+AceLexiconEntry, -ApeLexEntry)
%
% Converts a ACE Lexicon entry into one Apelex entry. This predicate succeeds several times
% if there are more than one Apelex entry for the ACE Lexicon entry.


% adverbs (old)
convert(AceLexiconEntry, ApeLexEntry) :-
	AceLexiconEntry = adv(
		adverb(Adverb),
		adverb_comparative(AdverbComparative),
		adverb_superlative(AdverbSuperlative),
		adverb_aliases(AdverbAliases),
		adverb_comparative_aliases(AdverbComparativeAliases),
		adverb_superlative_aliases(AdverbSuperlativeAliases),
		type(Type),
		comment(Comment)
	),
	token(Adverb),
	list_of_atoms([AdverbComparative, AdverbSuperlative, Type, Comment]),
	list_of_tokens(AdverbAliases),
	list_of_atoms(AdverbComparativeAliases),
	list_of_atoms(AdverbSuperlativeAliases),
	!,
	member(WordForm, [Adverb|AdverbAliases]),
	ApeLexEntry = adv(WordForm, Adverb, Type, Comment).


% adverbs (new)
convert(AceLexiconEntry, ApeLexEntry) :-
	AceLexiconEntry =.. [
		adv,
		positive(PositiveForms),
		comparative(ComparativeForms),
		superlative(SuperlativeForms)
		|Rest],
	tokens(PositiveForms, Positive, PositiveList),
	Positive \== '',  % at least one positive form must be present
	tokens(ComparativeForms, _Comparative, ComparativeList),
	tokens(SuperlativeForms, _Superlative, SuperlativeList),
	get_type_comment(Rest, Type, Comment),
	!,
	(
		% positive
		member(WordForm, PositiveList),
		ApeLexEntry = adv(WordForm, Positive, Type, Comment)
	;
		% comparative
		member(WordForm, ComparativeList),
		ApeLexEntry = adv_comp(WordForm, Positive, Type, Comment)
	;
		% superlative
		member(WordForm, SuperlativeList),
		ApeLexEntry = adv_sup(WordForm, Positive, Type, Comment)
	).


% intransitive adjective (old)
convert(AceLexiconEntry, ApeLexEntry) :-
	AceLexiconEntry = adj(
		positive(Positive),
		comparative(Comparative),
		superlative(Superlative),
		positive_aliases(PositiveAliases),
		comparative_aliases(ComparativeAliases),
		superlative_aliases(SuperlativeAliases),
		complementing_preposition(ComplementingPreposition),
		comment(Comment)
	),
	ComplementingPreposition == '',
	token_or_empty(Positive),
	atom(Comparative),
	atom(Superlative),
	list_of_tokens(PositiveAliases),
	list_of_atoms(ComparativeAliases),
	list_of_atoms(SuperlativeAliases),
	atom(Comment),
	!,
	(
		% positive
		Positive \= '',
		member(WordForm, [Positive|PositiveAliases]),
		ApeLexEntry = adj_itr(WordForm, Positive, na, Comment)
	;
		% comparative
		Comparative \= '',
		member(WordForm, [Comparative|ComparativeAliases]),
		ApeLexEntry = adj_itr_comp(WordForm, Positive, na, Comment)
	;
		% superlative
		Superlative \= '',
		member(WordForm, [Superlative|SuperlativeAliases]),
		ApeLexEntry = adj_itr_sup(WordForm, Positive, na, Comment)
	).


% intransitive adjective (new)
convert(AceLexiconEntry, ApeLexEntry) :-
	AceLexiconEntry =.. [
		adj,
		positive(PositiveForms),
		comparative(ComparativeForms),
		superlative(SuperlativeForms)
		|Rest],
	tokens(PositiveForms, Positive, PositiveList),
	Positive \== '',  % at least one positive form must be present
	tokens(ComparativeForms, _Comparative, ComparativeList),
	tokens(SuperlativeForms, _Superlative, SuperlativeList),
	get_type_comment(Rest, Type, Comment),
	!,
	(
		% positive
		member(WordForm, PositiveList),
		ApeLexEntry = adj_itr(WordForm, Positive, Type, Comment)
	;
		% comparative
		member(WordForm, ComparativeList),
		ApeLexEntry = adj_itr_comp(WordForm, Positive, Type, Comment)
	;
		% superlative
		member(WordForm, SuperlativeList),
		ApeLexEntry = adj_itr_sup(WordForm, Positive, Type, Comment)
	).


% transitive adjective (old)
convert(AceLexiconEntry, ApeLexEntry) :-
	AceLexiconEntry = adj(
		positive(Positive),
		comparative(Comparative),
		superlative(Superlative),
		positive_aliases(PositiveAliases),
		comparative_aliases(ComparativeAliases),
		superlative_aliases(SuperlativeAliases),
		complementing_preposition(ComplementingPreposition),
		comment(Comment)
	),
	token(Positive),
	token(ComplementingPreposition),
	list_of_atoms([Comparative, Superlative, Comment]),
	list_of_tokens(PositiveAliases),
	list_of_atoms(ComparativeAliases),
	list_of_atoms(SuperlativeAliases),
	!,
	hyphenate_list([Positive|PositiveAliases], ComplementingPreposition, [Positive2|PositiveAliases2]),
	hyphenate_list([Comparative|ComparativeAliases], ComplementingPreposition, ComparativeList2),
	hyphenate_list([Superlative|SuperlativeAliases], ComplementingPreposition, SuperlativeList2),
	(
		% positive
		member(WordForm, [Positive2|PositiveAliases2]),
		ApeLexEntry = adj_tr(WordForm, Positive2, ComplementingPreposition, na, Comment)
	;
		% comparative
		member(WordForm, ComparativeList2),
		ApeLexEntry = adj_tr_comp(WordForm, Positive2, ComplementingPreposition, na, Comment)
	;
		% superlative
		member(WordForm, SuperlativeList2),
		ApeLexEntry = adj_tr_sup(WordForm, Positive2, ComplementingPreposition, na, Comment)
	).


% transitive adjective (new)
convert(AceLexiconEntry, ApeLexEntry) :-
	AceLexiconEntry =.. [
		tr_adj,
		positive(PositiveForms),
		comparative(ComparativeForms),
		superlative(SuperlativeForms),
		preposition(Preposition)
		|Rest],
	tokens(PositiveForms, Positive, PositiveList),
	Positive \== '',  % at least one positive form must be present
	token(Preposition),
	tokens(ComparativeForms, _Comparative, ComparativeList),
	tokens(SuperlativeForms, _Superlative, SuperlativeList),
	get_type_comment(Rest, Type, Comment),
	!,
	(
		% positive
		member(WordForm, PositiveList),
		ApeLexEntry = adj_tr(WordForm, Positive, Preposition, Type, Comment)
	;
		% comparative
		member(WordForm, ComparativeList),
		ApeLexEntry = adj_tr_comp(WordForm, Positive, Preposition, Type, Comment)
	;
		% superlative
		member(WordForm, SuperlativeList),
		ApeLexEntry = adj_tr_sup(WordForm, Positive, Preposition, Type, Comment)
	).


% countable noun (old)
convert(AceLexiconEntry, ApeLexEntry) :-
	AceLexiconEntry = cn(
		singular(Singular),
		plural(Plural),
		singular_aliases(SingularAliases),
		plural_aliases(PluralAliases),
		type(Type),
		gender(Gender),
		group(Group),
		comment(Comment)
	),
	Group == countable,
	token_or_empty(Singular),
	token_or_empty(Plural),
	list_of_tokens(SingularAliases),
	list_of_tokens(PluralAliases),
	list_of_atoms([Type, Comment]),
	gender(Gender),
	!,
	get_lexem(Singular, Plural, Lexem),
	(
		% singular
		Singular \= '',
		member(WordForm, [Singular|SingularAliases]),
		ApeLexEntry = noun_sg(WordForm, Lexem, Gender, Type, Comment)
	;
		% plural
		Plural \= '',
		member(WordForm, [Plural|PluralAliases]),
		ApeLexEntry = noun_pl(WordForm, Lexem, Gender, Type, Comment)
	).


% countable noun (new)
convert(AceLexiconEntry, ApeLexEntry) :-
	AceLexiconEntry =.. [
		count_noun,
		singular(SingularForms),
		plural(PluralForms),
		gender(Gender)
		|Rest],
	tokens(SingularForms, Singular, SingularList),
	tokens(PluralForms, Plural, PluralList),
	gender(Gender),
	get_type_comment(Rest, Type, Comment),
	!,
	get_lexem(Singular, Plural, Lexem),
	(
		% singular
		member(WordForm, SingularList),
		ApeLexEntry = noun_sg(WordForm, Lexem, Gender, Type, Comment)
	;
		% plural
		member(WordForm, PluralList),
		ApeLexEntry = noun_pl(WordForm, Lexem, Gender, Type, Comment)
	).


% mass noun (old)
convert(AceLexiconEntry, ApeLexEntry) :-
	AceLexiconEntry = cn(
		singular(Singular),
		plural(Plural),
		singular_aliases(SingularAliases),
		plural_aliases(PluralAliases),
		type(Type),
		gender(Gender),
		group(Group),
		comment(Comment)
	),
	Group == mass,
	Plural == '',
	PluralAliases == [],
	token(Singular),
	list_of_tokens(SingularAliases),
	list_of_atoms([Type, Comment]),
	gender(Gender),
	!,
	member(WordForm, [Singular|SingularAliases]),
	ApeLexEntry = noun_mass(WordForm, Singular, Gender, Type, Comment).


% mass noun (new)
convert(AceLexiconEntry, ApeLexEntry) :-
	AceLexiconEntry =.. [
		mass_noun,
		singular(SingularForms)
		|Rest],
	tokens(SingularForms, Singular, SingularList),
	get_type_comment(Rest, Type, Comment),
	!,
	member(WordForm, SingularList),
	ApeLexEntry = noun_mass(WordForm, Singular, neutr, Type, Comment).


% measurement noun (old)
convert(AceLexiconEntry, ApeLexEntry) :-
	AceLexiconEntry = mn(
		singular(Singular),
		plural(Plural),
		singular_aliases(SingularAliases),
		plural_aliases(PluralAliases),
		dimension(Type),
		comment(Comment)
	),
	token_or_empty(Singular),
	token_or_empty(Plural),
	list_of_tokens(SingularAliases),
	list_of_tokens(PluralAliases),
	list_of_atoms([Type, Comment]),
	!,
	get_lexem(Singular, Plural, Lexem),
	(
		% singular
		Singular \= '',
		member(WordForm, [Singular|SingularAliases]),
		ApeLexEntry = mn_sg(WordForm, Lexem, Type, Comment)
	;
		% plural
		Plural \= '',
		member(WordForm, [Plural|PluralAliases]),
		ApeLexEntry = mn_pl(WordForm, Lexem, Type, Comment)
	).


% measurement noun (new)
convert(AceLexiconEntry, ApeLexEntry) :-
	AceLexiconEntry =.. [
		measure_noun,
		singular(SingularForms),
		plural(PluralForms)
		|Rest],
	tokens(SingularForms, Singular, SingularList),
	tokens(PluralForms, Plural, PluralList),
	get_type_comment(Rest, Type, Comment),
	!,
	get_lexem(Singular, Plural, Lexem),
	(
		% singular
		member(WordForm, SingularList),
		ApeLexEntry = mn_sg(WordForm, Lexem, Type, Comment)
	;
		% plural
		member(WordForm, PluralList),
		ApeLexEntry = mn_pl(WordForm, Lexem, Type, Comment)
	).


% proper name (old)
convert(AceLexiconEntry, ApeLexEntry) :-
	AceLexiconEntry = pn(
		singular(Singular),
		plural(Plural),
		singular_aliases(SingularAliases),
		plural_aliases(PluralAliases),
		type(Type),
		gender(Gender),
		comment(Comment)
	),
	token_or_empty(Singular),
	token_or_empty(Plural),
	list_of_tokens(SingularAliases),
	list_of_tokens(PluralAliases),
	list_of_atoms([Type, Comment]),
	gender(Gender),
	!,
	get_lexem(Singular, Plural, Lexem),
	(
		% singular
		Singular \= '',
		member(WordForm, [Singular|SingularAliases]),
		ApeLexEntry = pn_sg(WordForm, Lexem, Gender, Type, Comment)
	;
		% plural
		Plural \= '',
		member(WordForm, [Plural|PluralAliases]),
		ApeLexEntry = pn_pl(WordForm, Lexem, Gender, Type, Comment)
	).


% proper name (new)
convert(AceLexiconEntry, ApeLexEntry) :-
	AceLexiconEntry =.. [
		propername,
		name(ProperNameForms),
		number(Number),
		gender(Gender)
		|Rest],
	tokens(ProperNameForms, ProperName, ProperNameList),
	gender(Gender),
	atom(Number),
	member(Number, [singular, plural]),
	get_type_comment(Rest, Type, Comment),
	!,
	(
		% singular
		Number == singular,
		member(WordForm, ProperNameList),
		ApeLexEntry = pn_sg(WordForm, ProperName, Gender, Type, Comment)
	;
		% plural
		Number == plural,
		member(WordForm, ProperNameList),
		ApeLexEntry = pn_pl(WordForm, ProperName, Gender, Type, Comment)
	).


% intransitive verb (old)
convert(AceLexiconEntry, ApeLexEntry) :-
	AceLexiconEntry = iv(
		third_singular(ThirdSingular),
		third_plural(ThirdPlural),
		third_singular_aliases(ThirdSingularAliases),
		third_plural_aliases(ThirdPluralAliases),
		type(Type),
		phrasal_particle(PhrasalParticle),
		comment(Comment)
	),
	token_or_empty(ThirdSingular),
	token_or_empty(ThirdPlural),
	token_or_empty(PhrasalParticle),
	list_of_atoms([Type, Comment]),
	list_of_tokens(ThirdSingularAliases),
	list_of_tokens(ThirdPluralAliases),
	!,
	hyphenate_list([ThirdSingular|ThirdSingularAliases], PhrasalParticle, [ThirdSingular2|ThirdSingularAliases2]),
	hyphenate_list([ThirdPlural|ThirdPluralAliases], PhrasalParticle, [ThirdPlural2|ThirdPluralAliases2]),
	get_lexem(ThirdPlural2, ThirdSingular2, Lexem),
	(
		% singular
		ThirdSingular \= '',
		member(WordForm, [ThirdSingular2|ThirdSingularAliases2]),
		ApeLexEntry = iv_finsg(WordForm, Lexem, Type, Comment)
	;
		% plural
		ThirdPlural \= '',
		member(WordForm, [ThirdPlural2|ThirdPluralAliases2]),
		ApeLexEntry = iv_infpl(WordForm, Lexem, Type, Comment)
	).


% intransitive verb (new)
convert(AceLexiconEntry, ApeLexEntry) :-
	AceLexiconEntry =.. [
		intr_verb,
		third_singular(ThirdSingularForms),
		third_plural(ThirdPluralForms)
		|Rest],
	tokens(ThirdSingularForms, ThirdSingular, ThirdSingularList),
	tokens(ThirdPluralForms, ThirdPlural, ThirdPluralList),
	get_type_comment(Rest, Type, Comment),
	!,
	get_lexem(ThirdPlural, ThirdSingular, Lexem),
	(
		% singular
		member(WordForm, ThirdSingularList),
		ApeLexEntry = iv_finsg(WordForm, Lexem, Type, Comment)
	;
		% plural
		member(WordForm, ThirdPluralList),
		ApeLexEntry = iv_infpl(WordForm, Lexem, Type, Comment)
	).


% transitive verb (old)
convert(AceLexiconEntry, ApeLexEntry) :-
	AceLexiconEntry = tv(
		third_singular(ThirdSingular),
		third_plural(ThirdPlural),
		third_singular_aliases(ThirdSingularAliases),
		third_plural_aliases(ThirdPluralAliases),
		type(Type),
		phrasal_particle(PhrasalParticle),
		direct_preposition(DirectPreposition),
		comment(Comment)
	),
	token_or_empty(ThirdSingular),
	token_or_empty(ThirdPlural),
	token_or_empty(PhrasalParticle),
	token_or_empty(DirectPreposition),
	list_of_atoms([Type, Comment]),
	list_of_tokens(ThirdSingularAliases),
	list_of_tokens(ThirdPluralAliases),
	!,
	hyphenate_list([ThirdSingular|ThirdSingularAliases], PhrasalParticle, [ThirdSingular2|ThirdSingularAliases2]),
	hyphenate_list([ThirdSingular2|ThirdSingularAliases2], DirectPreposition, [ThirdSingular3|ThirdSingularAliases3]),
	hyphenate_list([ThirdPlural|ThirdPluralAliases], PhrasalParticle, [ThirdPlural2|ThirdPluralAliases2]),
	hyphenate_list([ThirdPlural2|ThirdPluralAliases2], DirectPreposition, [ThirdPlural3|ThirdPluralAliases3]),
	get_lexem(ThirdPlural3, ThirdSingular3, Lexem),
	(
		% singular
		ThirdSingular \= '',
		member(WordForm, [ThirdSingular3|ThirdSingularAliases3]),
		ApeLexEntry = tv_finsg(WordForm, Lexem, Type, Comment)
	;
		% plural
		ThirdPlural \= '',
		member(WordForm, [ThirdPlural3|ThirdPluralAliases3]),
		ApeLexEntry = tv_infpl(WordForm, Lexem, Type, Comment)
	).


% transitive verb (new)
convert(AceLexiconEntry, ApeLexEntry) :-
	AceLexiconEntry =.. [
		tr_verb,
		third_singular(ThirdSingularForms),
		third_plural(ThirdPluralForms),
		past_participle(PastParticipleForms)
		|Rest],
	tokens(ThirdSingularForms, ThirdSingular, ThirdSingularList),
	tokens(ThirdPluralForms, ThirdPlural, ThirdPluralList),
	tokens(PastParticipleForms, _PastParticiple, PastParticipleList),
	get_type_comment(Rest, Type, Comment),
	!,
	get_lexem(ThirdPlural, ThirdSingular, Lexem),
	(
		% singular
		member(WordForm, ThirdSingularList),
		ApeLexEntry = tv_finsg(WordForm, Lexem, Type, Comment)
	;
		% plural
		member(WordForm, ThirdPluralList),
		ApeLexEntry = tv_infpl(WordForm, Lexem, Type, Comment)
	;
		% past participle
		member(WordForm, PastParticipleList),
		ApeLexEntry = tv_pp(WordForm, Lexem, Type, Comment)
	).


% ditransitive verb (old)
convert(AceLexiconEntry, ApeLexEntry) :-
	AceLexiconEntry = dv(
		third_singular(ThirdSingular),
		third_plural(ThirdPlural),
		third_singular_aliases(ThirdSingularAliases),
		third_plural_aliases(ThirdPluralAliases),
		type(Type),
		phrasal_particle(PhrasalParticle),
		indirect_preposition(IndirectPreposition),
		allow_dative_shift(DativeShift),
		comment(Comment)
	),
	token_or_empty(ThirdSingular),
	token_or_empty(ThirdPlural),
	token_or_empty(PhrasalParticle),
	token_or_empty(IndirectPreposition),
	list_of_atoms([Type, DativeShift, Comment]),
	list_of_tokens(ThirdSingularAliases),
	list_of_tokens(ThirdPluralAliases),
	member(DativeShift, [yes, no]),
	!,
	hyphenate_list([ThirdSingular|ThirdSingularAliases], PhrasalParticle, [ThirdSingular2|ThirdSingularAliases2]),
	hyphenate_list([ThirdPlural|ThirdPluralAliases], PhrasalParticle, [ThirdPlural2|ThirdPluralAliases2]),
	get_lexem(ThirdPlural2, ThirdSingular2, Lexem),
	(
		% singular
		ThirdSingular \= '',
		member(WordForm, [ThirdSingular2|ThirdSingularAliases2]),
		(
			% without indirect preposition
			ditr_without_prep(DativeShift, IndirectPreposition),
			ApeLexEntry = dv_finsg(WordForm, Lexem, '', Type, Comment)
		;
			% with indirect preposition
			IndirectPreposition \== '',
			ApeLexEntry = dv_finsg(WordForm, Lexem, IndirectPreposition, Type, Comment)
		)
	;
		% plural
		ThirdPlural \= '',
		member(WordForm, [ThirdPlural2|ThirdPluralAliases2]),
		(
			% without indirect preposition
			ditr_without_prep(DativeShift, IndirectPreposition),
			ApeLexEntry = dv_infpl(WordForm, Lexem, '', Type, Comment)
		;
			% with indirect preposition
			IndirectPreposition \== '',
			ApeLexEntry = dv_infpl(WordForm, Lexem, IndirectPreposition, Type, Comment)
		)
	).


% ditransitive verb (new)
convert(AceLexiconEntry, ApeLexEntry) :-
	AceLexiconEntry =.. [
		ditr_verb,
		third_singular(ThirdSingularForms),
		third_plural(ThirdPluralForms),
		past_participle(PastParticipleForms),
		indirect_preposition(IndirectPreposition),
		allow_dative_shift(DativeShift)
		|Rest],
	tokens(ThirdSingularForms, ThirdSingular, ThirdSingularList),
	tokens(ThirdPluralForms, ThirdPlural, ThirdPluralList),
	tokens(PastParticipleForms, _PastParticiple, PastParticipleList),
	token_or_empty(IndirectPreposition),
	atom(DativeShift),
	member(DativeShift, [yes, no]),
	get_type_comment(Rest, Type, Comment),
	!,
	get_lexem(ThirdPlural, ThirdSingular, Lexem),
	(
		% singular
		member(WordForm, ThirdSingularList),
		(
			% without indirect preposition
			ditr_without_prep(DativeShift, IndirectPreposition),
			ApeLexEntry = dv_finsg(WordForm, Lexem, '', Type, Comment)
		;
			% with indirect preposition
			IndirectPreposition \== '',
			ApeLexEntry = dv_finsg(WordForm, Lexem, IndirectPreposition, Type, Comment)
		)
	;
		% plural
		member(WordForm, ThirdPluralList),
		(
			% without indirect preposition
			ditr_without_prep(DativeShift, IndirectPreposition),
			ApeLexEntry = dv_infpl(WordForm, Lexem, '', Type, Comment)
		;
			% with indirect preposition
			IndirectPreposition \== '',
			ApeLexEntry = dv_infpl(WordForm, Lexem, IndirectPreposition, Type, Comment)
		)
	;
		% past participle
		member(WordForm, PastParticipleList),
		(
			% without indirect preposition
			ditr_without_prep(DativeShift, IndirectPreposition),
			ApeLexEntry = dv_pp(WordForm, Lexem, '', Type, Comment)
		;
			% with indirect preposition
			IndirectPreposition \== '',
			ApeLexEntry = dv_pp(WordForm, Lexem, IndirectPreposition, Type, Comment)
		)
	).


%% get_type_comment(+RestList, -Type, -Comment)

get_type_comment([], na, '').

get_type_comment([type(Type)], Type, '') :-
    atom(Type).

get_type_comment([comment(Comment)], na, Comment) :-
    atom(Comment).

get_type_comment([type(Type), comment(Comment)], Type, Comment) :-
    atom(Type),
    atom(Comment).


%% ditr_without_prep(+DativeShift, +IndirectPreposition)
%
% A ditransitive verb can be used without an indirect preposition if (and only if)
% dative shift is allowed or no indirect preposition is defined.

ditr_without_prep(yes, _) :-
    !.

ditr_without_prep(_, '').


%% gender(+Gender)
%
% Succeeds if Gender is a valid gender atom.

gender(X) :-
    atom(X),
    member(X, [neutr, human, fem, masc]).


%% transform_spaces(+Atom, -NewAtom)
%
% Transforms the blank spaces in Atom into underscores.

transform_spaces(Atom, NewAtom) :-
	concat_atom(AtomList, ' ', Atom),
	concat_atom(AtomList, '_', NewAtom).


%% get_lexem(+Word1, +Word2, -Lexem)
%
% Transforms the hyphens of Word1 into underscores. If Word1 is
% empty ('') then Word2 is used.

get_lexem('', Word, Word) :-
	!.

get_lexem(Word, _, Word).


%% hyphenate_list(+WordListIn, +Word2, -WordListOut)
%
% Hyphenates Word2 to every element of the list WordListIn.

hyphenate_list([], _, []).

hyphenate_list([Word1|Rest], '', [Word1|RestOut]) :-
    !,
	hyphenate_list(Rest, '', RestOut).

hyphenate_list([Word1|Rest], Word2, [WordOut|RestOut]) :-
    concat_atom([Word1, Word2], '-', WordOut),
	hyphenate_list(Rest, Word2, RestOut).


%% list_of_atoms(+List)
%
% Succeeds if List is a list of atoms.

list_of_atoms(V) :-
    var(V),
    !,
    fail.

list_of_atoms([]).

list_of_atoms([H|T]) :-
	atom(H),
	list_of_atoms(T).


%% tokens(+In, -First, -List)
%
% In can be an atom or a list of atoms. First returns the first atom.
% List returns the list of all atoms. This predicate succeeds only if
% all atoms are valid tokens.

tokens('', '', []) :-
    !.

tokens([], '', []) :-
    !.

tokens([First|Rest], First, [First|Rest]) :-
    !,
    list_of_tokens([First|Rest]).

tokens(First, First, [First]) :-
    token(First).


%% list_of_tokens(+List)
%
% Succeeds if List is a list of valid tokens.

list_of_tokens([]).

list_of_tokens([H|T]) :-
	token(H),
	list_of_tokens(T).


%% token_or_empty(+Token)
%
% Succeeds if Token is either a valid token or an empty atom ''.

token_or_empty(Token) :-
    Token == ''.

token_or_empty(Token) :-
    token(Token).


%% token(+Token)
%
% Succeeds if Token is a valid token. A token is valid if (and only if)
% it starts with a letter and the following zero or more characters are
% letters, numbers, hyphens, or underscores.

token(Token) :-
    atom(Token),
    atom_codes(Token, [First|Rest]),
    is_letter(First),
    valid_token_code_list(Rest).


%% valid_token_code_list(+TokenList)
%
% Succeeds if the list contains only valid character codes.

valid_token_code_list([]).

valid_token_code_list([First|Rest]) :-
    is_letter(First),
    valid_token_code_list(Rest).

valid_token_code_list([First|Rest]) :-
    is_digit(First),
    valid_token_code_list(Rest).

valid_token_code_list([45|Rest]) :-  % hyphen
    valid_token_code_list(Rest).

valid_token_code_list([95|Rest]) :-  % underscore
    valid_token_code_list(Rest).


%% is_number(+Atom)
%
% Succeeds if Atom can be read as a number. E.g. it succeeds for '99' or '8.3'.

is_number(Token) :-
	catch( atom_number(Token, _), _, fail ).


%% is_symbol(+Atom)
%
% Succeds if Atom is a symbol that is accepted by the tokenizer.

is_symbol('.').
is_symbol('?').
is_symbol('!').
is_symbol(':').
is_symbol('\'').
is_symbol('/').
is_symbol(',').
