% This file is part of the Attempto Parsing Engine (APE).
% Copyright 2008, Attempto Group, University of Zurich (see http://attempto.ifi.uzh.ch).
%
% The Attempto Parsing Engine (APE) is free software: you can redistribute it and/or modify it
% under the terms of the GNU Lesser General Public License as published by the Free Software
% Foundation, either version 3 of the License, or (at your option) any later version.
%
% The Attempto Parsing Engine (APE) is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
% PURPOSE. See the GNU Lesser General Public License for more details.
%
% You should have received a copy of the GNU Lesser General Public License along with the Attempto
% Parsing Engine (APE). If not, see http://www.gnu.org/licenses/.


:- module(ulex, [
		read_ulex/1,            % +AceLexiconFile
		add_lexicon_entries/1,  % +LexEntryList
		add_lexicon_entry/1,    % +LexEntry
		discard_ulex/0
	]).

:- use_module(acelexicon_apelex).
:- use_module(functionwords).
:- use_module('../logger/error_logger').

/** <module> Dynamic User Lexicon

This module contains the predicates for the dynamic management of a user lexicon.

@author Tobias Kuhn
@version 2007-05-29
*/


% The Ulex predicates are declared dynamic.

:- dynamic adv/4.
:- dynamic adv_comp/4.
:- dynamic adv_sup/4.
:- dynamic adj_itr/4.
:- dynamic adj_itr_comp/4.
:- dynamic adj_itr_sup/4.
:- dynamic adj_tr/5.
:- dynamic adj_tr_comp/5.
:- dynamic adj_tr_sup/5.
:- dynamic noun_sg/5.
:- dynamic noun_pl/5.
:- dynamic noun_mass/5.
:- dynamic mn_sg/4.
:- dynamic mn_pl/4.
:- dynamic pn_sg/5.
:- dynamic pn_pl/5.
:- dynamic iv_finsg/4.
:- dynamic iv_infpl/4.
:- dynamic tv_finsg/4.
:- dynamic tv_infpl/4.
:- dynamic tv_pp/4.
:- dynamic dv_finsg/5.
:- dynamic dv_infpl/5.
:- dynamic dv_pp/5.


%% read_ulex(+AceLexiconFile)
%
% Reads the AceLexicon file and loads all the contained entries.

read_ulex(AceLexicon) :-
	catch(
		(
			see(AceLexicon),
			read_lexicon_entries
		),
		_CatchType,
		(
			with_output_to(atom(Message), format("The user lexicon file is not a valid Prolog file.", [])),
			add_error_message_once(lexicon, '', 'Malformed file.', Message)
		)
	),
	seen.

read_ulex(_) :-
	seen.


%% read_lexicon_entries
%
% Loads the lexicon entries reading from the current input device.

read_lexicon_entries :-
    read(LexEntry),
    ( LexEntry == end_of_file ->
    	true
    ;
    	add_lexicon_entry(LexEntry),
    	read_lexicon_entries
    ).


%% add_lexicon_entries(+LexEntryList)
%
% @param LexEntryList
%
% Adds all the lexicon entries of the list to the dynamic lexicon.

add_lexicon_entries([]).

add_lexicon_entries([LexEntry|Rest]) :-
    add_lexicon_entry(LexEntry),
    add_lexicon_entries(Rest).


%% add_lexicon_entry(+LexEntry)
%
% @param LexEntry
%
% Adds the lexicon entry to the dynamic lexicon.

add_lexicon_entry(LexEntry) :-
	acelexicon_apelex(LexEntry, ApeLexEntries),
	( ApeLexEntries = [] ->
		with_output_to(atom(Message), format("User lexicon entry is malformed: ~w", [LexEntry])),
		add_error_message_once(lexicon, '', 'Malformed entry.', Message)
	;
		assert_lexicon_entries(ApeLexEntries)
	).


%% assert_lexicon_entries(+Entries)
%
% Assert the lexicon entries.

assert_lexicon_entries([]).

assert_lexicon_entries([ApeLexEntry|Rest]) :-
    check_intersections(ApeLexEntry),
    assert(ApeLexEntry),
    % uncommenting the next line outputs the Apelex entries
    %format(user_error, '~q.\n', [ApeLexEntry]),
    assert_lexicon_entries(Rest).


%% discard_ulex
%
% Discards all the lexicon entries.

discard_ulex :-
	retractall(adv(_,_,_,_)),
	retractall(adv_comp(_,_,_,_)),
	retractall(adv_sup(_,_,_,_)),
	retractall(adj_itr(_,_,_,_)),
	retractall(adj_itr_comp(_,_,_,_)),
	retractall(adj_itr_sup(_,_,_,_)),
	retractall(adj_tr(_,_,_,_,_)),
	retractall(adj_tr_comp(_,_,_,_,_)),
	retractall(adj_tr_sup(_,_,_,_,_)),
	retractall(noun_sg(_,_,_,_,_)),
	retractall(noun_pl(_,_,_,_,_)),
	retractall(noun_mass(_,_,_,_,_)),
	retractall(mn_sg(_,_,_,_)),
	retractall(mn_pl(_,_,_,_)),
	retractall(pn_sg(_,_,_,_,_)),
	retractall(pn_pl(_,_,_,_,_)),
	retractall(iv_finsg(_,_,_,_)),
	retractall(iv_infpl(_,_,_,_)),
	retractall(tv_finsg(_,_,_,_)),
	retractall(tv_infpl(_,_,_,_)),
	retractall(tv_pp(_,_,_,_)),
	retractall(dv_finsg(_,_,_,_,_)),
	retractall(dv_infpl(_,_,_,_,_)),
	retractall(dv_pp(_,_,_,_,_)).


%% check_intersection(+Entry)
%
% Checks if the new entry does not lead to a conflict. There is a conflict if
% - the word is a function word, or
% - the word is duplicated in the user lexicon, or
% - there is a harmful word class intersection.

check_intersections(Entry) :-
    Entry =.. [_,Word|_],
    unredefinable_fw(Word),
    add_error_message(lexicon, '', Word, 'This function word cannot be redefined in the user lexicon.').

check_intersections(Entry) :-
    Entry =.. [_,Word|_],
    functionwords:preposition(Word, _, _),
    add_error_message(lexicon, '', Word, 'This preposition cannot be redefined in the user lexicon.').

check_intersections(adv(Word,_,_,_)) :-
    ( adv(Word,_,_,_) ; adv_comp(Word,_,_,_) ; adv_sup(Word,_,_,_) ),
    add_error_message(lexicon, '', Word, 'This adverb is defined twice.').

check_intersections(adv_comp(Word,_,_,_)) :-
    ( adv(Word,_,_,_) ; adv_comp(Word,_,_,_) ; adv_sup(Word,_,_,_) ),
    add_error_message(lexicon, '', Word, 'This adverb is defined twice.').

check_intersections(adv_sup(Word,_,_,_)) :-
    ( adv(Word,_,_,_) ; adv_comp(Word,_,_,_) ; adv_sup(Word,_,_,_) ),
    add_error_message(lexicon, '', Word, 'This adverb is defined twice.').

check_intersections(adj_itr(Word,_,_,_)) :-
    ( adj_itr(Word,_,_,_) ; adj_itr_comp(Word,_,_,_) ; adj_itr_sup(Word,_,_,_) ),
    add_error_message(lexicon, '', Word, 'This intransitive adjective is defined twice.').

check_intersections(adj_itr_comp(Word,_,_,_)) :-
    ( adj_itr(Word,_,_,_) ; adj_itr_comp(Word,_,_,_) ; adj_itr_sup(Word,_,_,_) ),
    add_error_message(lexicon, '', Word, 'This intransitive adjective is defined twice.').

check_intersections(adj_itr_sup(Word,_,_,_)) :-
    ( adj_itr(Word,_,_,_) ; adj_itr_comp(Word,_,_,_) ; adj_itr_sup(Word,_,_,_) ),
    add_error_message(lexicon, '', Word, 'This intransitive adjective is defined twice.').

check_intersections(adj_tr(Word,_,_,_,_)) :-
    ( adj_tr(Word,_,_,_,_) ; adj_tr_comp(Word,_,_,_,_) ; adj_tr_sup(Word,_,_,_,_) ),
    add_error_message(lexicon, '', Word, 'This transitive adjective is defined twice.').

check_intersections(adj_tr_comp(Word,_,_,_,_)) :-
    ( adj_tr(Word,_,_,_,_) ; adj_tr_comp(Word,_,_,_,_) ; adj_tr_sup(Word,_,_,_,_) ),
    add_error_message(lexicon, '', Word, 'This transitive adjective is defined twice.').

check_intersections(adj_tr_sup(Word,_,_,_,_)) :-
    ( adj_tr(Word,_,_,_,_) ; adj_tr_comp(Word,_,_,_,_) ; adj_tr_sup(Word,_,_,_,_) ),
    add_error_message(lexicon, '', Word, 'This transitive adjective is defined twice.').

check_intersections(noun_sg(Word,_,_,_,_)) :-
    noun_sg(Word,_,_,_,_),
    add_error_message(lexicon, '', Word, 'This singular noun is defined twice.').

check_intersections(noun_pl(Word,_,_,_,_)) :-
    noun_pl(Word,_,_,_,_),
    add_error_message(lexicon, '', Word, 'This plural noun is defined twice.').

check_intersections(noun_mass(Word,_,_,_,_)) :-
    noun_mass(Word,_,_,_,_),
    add_error_message(lexicon, '', Word, 'This mass noun is defined twice.').

check_intersections(mn_sg(Word,_,_,_)) :-
    mn_sg(Word,_,_,_),
    add_error_message(lexicon, '', Word, 'This singular measurement noun is defined twice.').

check_intersections(mn_pl(Word,_,_,_)) :-
    mn_pl(Word,_,_,_),
    add_error_message(lexicon, '', Word, 'This plural measurement noun is defined twice.').

check_intersections(pn_sg(Word,_,_,_,_)) :-
    ( pn_sg(Word,_,_,_,_) ; pn_pl(Word,_,_,_,_) ),
    add_error_message(lexicon, '', Word, 'This proper name is defined twice.').

check_intersections(pn_pl(Word,_,_,_,_)) :-
    ( pn_sg(Word,_,_,_,_) ; pn_pl(Word,_,_,_,_) ),
    add_error_message(lexicon, '', Word, 'This proper name is defined twice.').

check_intersections(iv_finsg(Word,_,_,_)) :-
    iv_finsg(Word,_,_,_),
    add_error_message(lexicon, '', Word, 'This singular form of an intransitive verb is defined twice.').

check_intersections(iv_infpl(Word,_,_,_)) :-
    iv_infpl(Word,_,_,_),
    add_error_message(lexicon, '', Word, 'This plural form of an intransitive verb is defined twice.').

check_intersections(tv_finsg(Word,_,_,_)) :-
    tv_finsg(Word,_,_,_),
    add_error_message(lexicon, '', Word, 'This singular form of a transitive verb is defined twice.').

check_intersections(tv_infpl(Word,_,_,_)) :-
    tv_infpl(Word,_,_,_),
    add_error_message(lexicon, '', Word, 'This plural form of a transitive verb is defined twice.').

check_intersections(tv_pp(Word,_,_,_)) :-
    tv_pp(Word,_,_,_),
    add_error_message(lexicon, '', Word, 'This past participle form of a transitive verb is defined twice.').

check_intersections(dv_finsg(Word,_,'',_,_)) :-
    dv_finsg(Word,_,'',_,_),
    add_error_message(lexicon, '', Word, 'This singular form of a ditransitive verb is defined twice.').

check_intersections(dv_infpl(Word,_,'',_,_)) :-
    dv_infpl(Word,_,'',_,_),
    add_error_message(lexicon, '', Word, 'This plural form of a ditransitive verb is defined twice.').

check_intersections(dv_pp(Word,_,'',_,_)) :-
    dv_pp(Word,_,'',_,_),
    add_error_message(lexicon, '', Word, 'This past participle form of a ditransitive verb is defined twice.').

check_intersections(dv_finsg(Word,_,Prep,_,_)) :-
    dv_finsg(Word,_,Prep,_,_),
    Prep \== '',
    add_error_message(lexicon, '', Word, 'This singular form of a ditransitive verb is defined twice.').

check_intersections(dv_infpl(Word,_,Prep,_,_)) :-
    dv_infpl(Word,_,Prep,_,_),
    Prep \== '',
    add_error_message(lexicon, '', Word, 'This plural form of a ditransitive verb is defined twice.').

check_intersections(dv_pp(Word,_,Prep,_,_)) :-
    dv_pp(Word,_,Prep,_,_),
    Prep \== '',
    add_error_message(lexicon, '', Word, 'This past participle form of a ditransitive verb is defined twice.').

check_intersections(adv(Word,_,_,_)) :-
    noun_sg(Word,_,_,_,_),
    add_error_message(lexicon, '', Word, 'Bad intersection: adverb and singular noun.').

check_intersections(noun_sg(Word,_,_,_,_)) :-
    adv(Word,_,_,_),
    add_error_message(lexicon, '', Word, 'Bad intersection: adverb and singular noun.').

check_intersections(adv(Word,_,_,_)) :-
    noun_pl(Word,_,_,_,_),
    add_error_message(lexicon, '', Word, 'Bad intersection: adverb and plural noun.').

check_intersections(noun_pl(Word,_,_,_,_)) :-
    adv(Word,_,_,_),
    add_error_message(lexicon, '', Word, 'Bad intersection: adverb and plural noun.').

check_intersections(adv(Word,_,_,_)) :-
    noun_mass(Word,_,_,_,_),
    add_error_message(lexicon, '', Word, 'Bad intersection: adverb and mass noun.').

check_intersections(noun_mass(Word,_,_,_,_)) :-
    adv(Word,_,_,_),
    add_error_message(lexicon, '', Word, 'Bad intersection: adverb and mass noun.').

check_intersections(adv(Word,_,_,_)) :-
    iv_finsg(Word,_,_,_),
    add_error_message(lexicon, '', Word, 'Bad intersection: adverb and singular intransitive verb.').

check_intersections(iv_finsg(Word,_,_,_)) :-
    adv(Word,_,_,_),
    add_error_message(lexicon, '', Word, 'Bad intersection: adverb and singular intransitive verb.').

check_intersections(adv(Word,_,_,_)) :-
    iv_infpl(Word,_,_,_),
    add_error_message(lexicon, '', Word, 'Bad intersection: adverb and plural intransitive verb.').

check_intersections(iv_infpl(Word,_,_,_)) :-
    adv(Word,_,_,_),
    add_error_message(lexicon, '', Word, 'Bad intersection: adverb and plural intransitive verb.').

check_intersections(adj_itr(Word,_,_,_)) :-
    ( adj_tr(Word,_,_,_,_) ; adj_tr_comp(Word,_,_,_,_); adj_tr_sup(Word,_,_,_,_) ),
    add_error_message(lexicon, '', Word, 'Bad intersection: intransitive adjective and transitive adjective.').

check_intersections(adj_itr_comp(Word,_,_,_)) :-
    ( adj_tr(Word,_,_,_,_) ; adj_tr_comp(Word,_,_,_,_); adj_tr_sup(Word,_,_,_,_) ),
    add_error_message(lexicon, '', Word, 'Bad intersection: intransitive adjective and transitive adjective.').

check_intersections(adj_itr_sup(Word,_,_,_)) :-
    ( adj_tr(Word,_,_,_,_) ; adj_tr_comp(Word,_,_,_,_); adj_tr_sup(Word,_,_,_,_) ),
    add_error_message(lexicon, '', Word, 'Bad intersection: intransitive adjective and transitive adjective.').

check_intersections(adj_tr(Word,_,_,_,_)) :-
    ( adj_itr(Word,_,_,_) ; adj_itr_comp(Word,_,_,_) ; adj_itr_sup(Word,_,_,_) ),
    add_error_message(lexicon, '', Word, 'Bad intersection: intransitive adjective and transitive adjective.').

check_intersections(adj_tr_comp(Word,_,_,_,_)) :-
    ( adj_itr(Word,_,_,_) ; adj_itr_comp(Word,_,_,_) ; adj_itr_sup(Word,_,_,_) ),
    add_error_message(lexicon, '', Word, 'Bad intersection: intransitive adjective and transitive adjective.').

check_intersections(adj_tr_sup(Word,_,_,_,_)) :-
    ( adj_itr(Word,_,_,_) ; adj_itr_comp(Word,_,_,_) ; adj_itr_sup(Word,_,_,_) ),
    add_error_message(lexicon, '', Word, 'Bad intersection: intransitive adjective and transitive adjective.').

check_intersections(_).


unredefinable_fw(null).
unredefinable_fw(zero).
unredefinable_fw(one).
unredefinable_fw(two).
unredefinable_fw(three).
unredefinable_fw(four).
unredefinable_fw(five).
unredefinable_fw(six).
unredefinable_fw(seven).
unredefinable_fw(eight).
unredefinable_fw(nine).
unredefinable_fw(ten).
unredefinable_fw(eleven).
unredefinable_fw(twelve).
unredefinable_fw(dozen).
unredefinable_fw('Null').
unredefinable_fw('Zero').
unredefinable_fw('One').
unredefinable_fw('Two').
unredefinable_fw('Three').
unredefinable_fw('Four').
unredefinable_fw('Five').
unredefinable_fw('Six').
unredefinable_fw('Seven').
unredefinable_fw('Eight').
unredefinable_fw('Nine').
unredefinable_fw('Ten').
unredefinable_fw('Eleven').
unredefinable_fw('Twelve').
unredefinable_fw('Dozen').
unredefinable_fw('There').
unredefinable_fw(there).
unredefinable_fw(and).
unredefinable_fw(or).
unredefinable_fw(not).
unredefinable_fw(that).
unredefinable_fw(than).
unredefinable_fw(of).
unredefinable_fw('If').
unredefinable_fw(if).
unredefinable_fw(then).
unredefinable_fw(such).
unredefinable_fw(be).
unredefinable_fw('_that_').
unredefinable_fw(provably).
unredefinable_fw(more).
unredefinable_fw(most).
unredefinable_fw(least).
unredefinable_fw(less).
unredefinable_fw(are).
unredefinable_fw('Are').
unredefinable_fw(is).
unredefinable_fw('Is').
unredefinable_fw(the).
unredefinable_fw('The').
unredefinable_fw(a).
unredefinable_fw('A').
unredefinable_fw(an).
unredefinable_fw('An').
unredefinable_fw(some).
unredefinable_fw('Some').
unredefinable_fw(no).
unredefinable_fw('No').
unredefinable_fw(every).
unredefinable_fw('Every').
unredefinable_fw(all).
unredefinable_fw('All').
unredefinable_fw(each).
unredefinable_fw('Each').
unredefinable_fw(which).
unredefinable_fw('Which').
unredefinable_fw(its).
unredefinable_fw('Its').
unredefinable_fw(his).
unredefinable_fw('His').
unredefinable_fw(her).
unredefinable_fw('Her').
unredefinable_fw(their).
unredefinable_fw('Their').
unredefinable_fw(whose).
unredefinable_fw(it).
unredefinable_fw('It').
unredefinable_fw(he).
unredefinable_fw('He').
unredefinable_fw(she).
unredefinable_fw('She').
unredefinable_fw(they).
unredefinable_fw('They').
unredefinable_fw(him).
unredefinable_fw(them).
unredefinable_fw(itself).
unredefinable_fw(himself).
unredefinable_fw(herself).
unredefinable_fw(themselves).
unredefinable_fw(someone).
unredefinable_fw('Someone').
unredefinable_fw(somebody).
unredefinable_fw('Somebody').
unredefinable_fw(something).
unredefinable_fw('Something').
unredefinable_fw(nobody).
unredefinable_fw('Nobody').
unredefinable_fw(nothing).
unredefinable_fw('Nothing').
unredefinable_fw(everyone).
unredefinable_fw('Everyone').
unredefinable_fw(everybody).
unredefinable_fw('Everybody').
unredefinable_fw(everything).
unredefinable_fw('Everything').
unredefinable_fw(what).
unredefinable_fw('What').
unredefinable_fw(who).
unredefinable_fw('Who').
unredefinable_fw(how).
unredefinable_fw('How').
unredefinable_fw(where).
unredefinable_fw('Where').
unredefinable_fw(when).
unredefinable_fw('When').

