% This file is part of the Attempto Parsing Engine (APE).
% Copyright 2008, Attempto Group, University of Zurich (see http://attempto.ifi.uzh.ch).
%
% The Attempto Parsing Engine (APE) is free software: you can redistribute it and/or modify it
% under the terms of the GNU Lesser General Public License as published by the Free Software
% Foundation, either version 3 of the License, or (at your option) any later version.
%
% The Attempto Parsing Engine (APE) is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
% PURPOSE. See the GNU Lesser General Public License for more details.
%
% You should have received a copy of the GNU Lesser General Public License along with the Attempto
% Parsing Engine (APE). If not, see http://www.gnu.org/licenses/.


:- module(drs_to_coreace, [
		drs_to_coreace/2
	]).

/** <module> DRS to Core ACE verbalizer

Translate an Attempto DRS into Attempto Controlled English (ACE).
The target of this translation is not full ACE but a subset called Core ACE.

@author Kaarel Kaljurand
@version 2008-03-17

==
Usage:

drs_to_coreace:t("John likes Mary. A man likes Bill.").
drs_to_coreace:t("Every man waits.").
drs_to_coreace:t("What does a man like?").
drs_to_coreace:t("What does a man not like?").
drs_to_coreace:t("Where does a man not wait?").
drs_to_coreace:t("Which man waits?").
drs_to_coreace:t("John eats an n:apelsiin.").
drs_to_coreace:t("There are a man and a woman. The man waits.").
drs_to_coreace:t("Somebody sees himself.").
drs_to_coreace:t("5 men run.").
drs_to_coreace:t("A dog and a cat run.").
drs_to_coreace:t("There are a dog and a cat.").
drs_to_coreace:t("If there are a dog and a cat then a man waits.").
drs_to_coreace:t("Everything runs.").
drs_to_coreace:t("Each of 2 girls lifts herself.").
drs_to_coreace:t("Every man waits.", Paraphrase).

Those that are rejected currently but that could be supported:

?- drs_to_coreace:t("There are less than 3 men and less than 3 women.").
?- drs_to_coreace:t("Less than 3 men and John wait.").

For the developer: to see the internals, use:

?- drs_to_coreace:ref2conds(Referent, Conditions, BoxId, SentenceId).
==

The setup: we have DRS-boxes and discourse referents, such that

* Each referent lives-in exactly 1 box.
* For each box at least 1 referent lives-in it.
* Each referent has a unique ID and maps-to a set of conditions.
* Each box has a unique ID and has a name (if, then, '', or, `it is false that', etc.)

The task: verbalize each discourse referent, i.e. the conditions
of each discourse referent. And do it in the context of

1. the box,
2. whether the referent has been verbalized already,
3. whether some other referent from the same box has been verbalized already.

The order is important. It is the depth-first tree order on boxes (i.e. parent, children, neighbors).

TODO:

* Fix scope problems (maybe in APE):
** A customer enters every card or enters every code.

* Fix problem with interaction between anaphoric references and relative clauses.

* ??? Sometimes use Reflexive pronouns also when definite NPs are available: himself, herself, itself.

* ??? Paraphrase should reflect the original sentence order.

* Support queries.
** What does every man like?
** What does no man like?

* Support for-each-of plural constructions.

* Support for mixing 'and' and 'or' is buggy.
- TEST: Every pet is a dog or is a cat ,and is not a cow.

*/


:- use_module(typeddrs_to_untypeddrs, [
	typeddrs_to_untypeddrs/2
	]).

:- use_module(morphgen, [
	clear_vars/0,
	add_var/1,
	remove_singletons/2,
	listlist_listatom/2,
	format_dr/2,
	surface_noun/4,
	surface_verb/3,
	surface_property/3,
	surface_property/6,
	surface_adverb/3,
	surface_quotedstring/2
	]).

:- use_module(drs_utils, [
	get_toplevel_object_referents/2
	]).

/*
:- debug(verbose).
:- debug(that).
:- debug(toplevel).
:- debug(npcoord).
*/


% Operators used in the DRS
:- op(400, fy, -).
:- op(400, fy, ~).
:- op(400, fy, <>).
:- op(400, fy, []).
:- op(500, xfx, =>).
:- op(500, xfx, v).
:- op(500, xfx, '--').


:- dynamic is_rep/1, ref2conds/4, boxid/1.


%% drs_to_coreace(+Drs:drs, -AceSentenceList:list) is det.
%
% @param Drs is an Attempto DRS (untyped)
% @param AceSentenceList is a verbalization of the input DRS in Core ACE (a list of Core ACE sentences)
%
%==
% Empty DRS -> []
% Unsupported DRS -> []
% Supported DRS -> [sentence1, sentence2, ..., sentenceN], where sentence is an atom
%==
%
drs_to_coreace(drs([], []), []) :- !.

drs_to_coreace(DRS, AceSentenceList) :-
	retractall(is_rep(_)),
	clear_vars,
	retractall(morphgen:name_namespace(_, _)),
	assert(is_rep(nil)),

	% Make a copy to preserve the variables in the caller.
	copy_term(DRS, DRSCopy),

	DRSCopy = drs(_, ConditionList),
	drs_utils:get_toplevel_object_referents(ConditionList, toplevel(ToplevelReferentList, UnsortedSubjectList, UnsortedObjectList, NamedList)),

	debug(toplevel, "toplevel referents:
		ToplevelReferentList: ~w
		UnsortedSubjectList: ~w
		UnsortedObjectList: ~w
		NamedList: ~w~n",
		[ToplevelReferentList, UnsortedSubjectList, UnsortedObjectList, NamedList]),

	numbervars(DRSCopy, 0, _),
	drs_to_coreace_nvar(DRSCopy, [ToplevelReferentList, UnsortedSubjectList, UnsortedObjectList, NamedList], []--_, Result1),
	!,
	remove_singletons(Result1, Result2),
	find_sentencelist(Result2, AceTokenList),
	debug(verbose, 'Result:~n~w~nWithout singletons: ~w~nSentenceList: ~w~n', [Result1, Result2, AceTokenList]),
	listlist_listatom(AceTokenList, AceSentenceList),
	!.

drs_to_coreace(_, []).


%% drs_to_coreace_nvar(+Drs:drs, +Toplevel:list, -AceList:list) is det.
%
% @param Drs is an Attempto DRS (untyped) with variables nvard
% @param Toplevel is a list of 4 lists containing referents of various types
% @param AceText is a verbalization of the DRS in Core ACE
%
drs_to_coreace_nvar(Drs, [ToplevelReferentList, UnsortedSubjectList, UnsortedObjectList, NamedList], RI--RO, Result1) :-
	make_ref2conds(Drs),

	% The next line means, that we use the `there is' construction
	% only if needed. If there is a verb which calls a noun
	% then the noun will not be verbalized by `there is'.
	findall(Referent, use_there_is(Referent), UncalledReferents),

	% Remove syntactic objects because we do not want to use 'there is' with objects.
	% E.g. There is a cat. John hates the cat. --> John hates a cat.
	subtract(ToplevelReferentList, UnsortedObjectList, List1),

	% Add syntactic subjects because we want to use 'there is' with subjects.
	% E.g. A cat sees John. --> There is a cat X1. The cat X1 sees John.
	append(List1, UnsortedSubjectList, List2),

	% Remove named-objects to avoid e.g. "There is John.", regardless
	% of whether `John' is later used as a subject or as an object.
	subtract(List2, NamedList, UnnamedList2),

	% Remove duplicates
	list_to_set(UnnamedList2, InitialReferents),

	debug(verbose, "InitialReferents: ~w~n", [InitialReferents]),

	append(InitialReferents, UncalledReferents, List3),
	list_to_set(List3, List4),

	findall(SentenceId-Referent, (member(Referent, List4), ref2conds(Referent, _, _, SentenceId)), List5),
	keysort(List5, List6),
	findall(Referent, member(_SentenceId-Referent, List6), List7),

	debug(verbose, "input to call_referents_to_acetexts/2: ~w~n", [List7]),

	call_referents_to_acetexts(List7, RI--RO, Result1).


%% find_sentencelist(+TokenList:list, -SentenceList:list) is det.
%
% @param TokenList is a list of ACE tokens
% @param SentenceList is a list of sentences (where a sentence is a list of ACE tokens)
%
find_sentencelist([], []).

find_sentencelist(TokenList, [Sentence | SentenceList]) :-
	find_sentence(TokenList, Sentence, RestTokenList),
	find_sentencelist(RestTokenList, SentenceList).


%% find_sentence(+TokenList:list, -Sentence:list, -RestTokenList:list) is det.
%
% @param TokenList is a list of ACE tokens
% @param Sentence is a list of sentences (where a sentence is a list of ACE tokens)
% @param RestTokenList is a list of ACE tokens
%
find_sentence(['.' | RestTokenList], [], RestTokenList) :- !.

find_sentence([Token | RestTokenList], [Token | RestSentence], RestTokenList2) :-
	find_sentence(RestTokenList, RestSentence, RestTokenList2).


%% use_there_is(+Referent:nvar) is det.
%
% @param Referent is a discourse referent
%
% Succeeds if the referent is a toplevel referent.
%
% Referent is toplevel if:
%
% 1. Its condition list is not empty. (BUG: redundant?)
% 2. It is not called by any other Referent in the same box.
%
use_there_is(Referent) :-
	ref2conds(Referent, [Condition | ConditionList], BoxId, _),
	\+ is_called(Referent, [Condition | ConditionList], BoxId).


%% is_called(+Referent:nvar, +BoxId:term) is det.
%
% @param Referent is a discourse referent
% @param BoxId is a DRS box identifier
%
% Succeeds if Referent corresponds to a propername (because proper names are always called).
% Succeeds if Referent is used as a non-first argument (i.e. "called") in any condition in the DRS.
%
is_called(Referent, ReferentConditions, _) :-
	member(object(Referent, _, named, _, _, _), ReferentConditions),
	!.

is_called(Referent, _, BoxId) :-
	ref2conds(Referent2, ConditionList, BoxId, _),
	Referent \= Referent2,
	member(Condition, ConditionList),
	Condition \= formula(_, _, _),
	arg(Num, Condition, Referent),
	Num > 1.


%% call_referents_to_acetext(+RefList:list, -Result:list) is det.
%
% @param RefList is a list of top-level DRS referents
% @param Result is an ACE text (list of tokens)
%
call_referents_to_acetexts(RefList, RI--RO, Result) :-
	(
		referents_to_acetexts(RefList, RI--RO, ACETexts)
	->
		flatten(ACETexts, Result)
	;
		Result = []
	).


%% referents_to_acetexts(+RefList:list, -AceText:list) is det.
%
% @param RefList is a list of top-level DRS referents
% @param AceText is an ACE text (list of tokens)
%
referents_to_acetexts([], RI--RI, '.').

referents_to_acetexts([Referent | ReferentsTail], RI--RO, [Text | TextTail]) :-
	debug(verbose, "old referents: ~w~n", [RI]),
	referent_text(Referent, top, RI--RTmp, Text, _),
	debug(verbose, "top referent: ~w, text: ~w, old referents: ~w~n", [Referent, Text, RTmp]),
	!,
	referents_to_acetexts(ReferentsTail, RTmp--RO, TextTail).


%% referent_text(+Referent:nvar, -Text:list, -Features:list) is det.
%% referent_text(+Referent:nvar, +TopDeep:atom, -Text:list, -Features:list) is det.
%
% @param Referent is a discourse referent or data item
% @param TopDeep is in {top, deep}
% @param Text is an ACE text that corresponds to the discourse referent
% @param Features is in {f(sg, subj), f(sg, obj), f(pl, subj), f(pl, obj)}, for number and case sharing
%
referent_text(Referent, RI--RO, Text, Features) :-
	referent_text(Referent, deep, RI--RO, Text, Features).


referent_text(Referent, deep, RI--RO, [that, SentenceListCoord], _) :-
	ref2conds(Referent, [drs(Dom, CondList)], _BoxId, _),
	debug(that, "that: drs: ~w~n", [drs(Dom, CondList)]),
	drs_to_coreace_nvar(drs(Dom, CondList), [[], [], [], []], RI--RO, Result),
	find_sentencelist(Result, [_Ignore_BUG | SentenceList]),
	debug(that, "that: sentence(s): ~w~n", [SentenceList]),
	sentencelist_sentencelistcoord(SentenceList, SentenceListCoord).

referent_text(formula(Expr1, Eq, Expr2), top, RI--RO, [BoxMarker, Expr1Text, Eq, Expr2Text], _) :-
	!,
	expr_exprpp(Expr1, RI--RTmp, Expr1Text),
	expr_exprpp(Expr2, RTmp--RO, Expr2Text),
	ref2conds(formula(Expr1, Eq, Expr2), _, BoxId, _),
	get_box_prefix(BoxId, BoxMarker).

% Support for query pronouns: 'who' and 'what'.
% Note: 'how' and 'which' are not covered by this rule.
% Note: the question mark is added later.
% Note: we only support top-level ([0, top, []]) query-conditions.
% Note: this rule generates "there is who?" and never "there are who?"
referent_text(Referent, TopDeep, RI--[Referent | RI], [ThereIs, qp(QueryPronoun)], _Features) :-
	ref2conds(Referent, [query(Referent, QueryPronoun)], [0, top, []], _),
	get_box_prefix([0, top, []], BoxPrefix),
	get_thereis(TopDeep, sg, BoxPrefix, ThereIs),
	\+ memberchk(Referent, RI).

referent_text(Referent, TopDeep, RI--RO, Text, Features) :-
	get_type(Referent, Type, BoxId),
	get_box_prefix(BoxId, BoxPrefix),
	debug(that, 'referent_text: ~W ~w ~w ~w~n', [Referent, [numbervars(true)], Type, BoxId, BoxPrefix]),
	!,
	conds_text(Type, BoxPrefix, TopDeep, RI--RO, Text, Features),
	debug(verbose, "referent_text_out: ~w; referents: ~w~n", [Text, RO]).

referent_text(Expression, deep, RI--RO, ExpressionText, f(sg, _)) :-
		expr_exprpp(Expression, RI--RO, ExpressionText).

% BUG: kaarel: 060804: fallback just in case
referent_text(_, _, RI-RI, '???', _).


%% get_type(+Referent:nvar, -Type:term, -BoxId:term) is det.
%
% @param Referent is a discourse referent
% @param Type is a structured form of Conditions
% @param BoxId is an identifier of the DRS box
%
% Type is a structured representation of the conditions, e.g. we group
% all properties and modifiers together into a list and represent as
% one argument.

get_type(Referent, Type, BoxId) :-
	ref2conds(Referent, Conditions, BoxId, _),
	conditions_type(Conditions, Type).


%% conditions_type(+Conditions:list, -Type:term) is det.
%
% @param Conditions is a list of DRS conditions that correspond to a referent
% @param Type is a structured form of those conditions
%
% There are three types of referents:
%
% 1. Those that have an object-condition
% 2. Those that have a predicate-condition
% 3. Those that have no conditions (i.e. the that-subordination referent)

conditions_type(Conditions, noun(Mains, parts(Parts), owners(Owners), adjectives(Adjectives))) :-
	member(Condition, Conditions),
	functor(Condition, object, _),
	!,
	conditions_nountype(Conditions, Mains, Parts, Owners, Adjectives).

conditions_type(Conditions, verb(Predicates, adverbs(Adverbs), pps(PPs))) :-
	member(Condition, Conditions),
	functor(Condition, predicate, _),
	!,
	conditions_verbtype(Conditions, Predicates, Adverbs, PPs).

conditions_type(Conditions, Conditions).


%% conditions_nountype(+Conditions:list, -Mains:list, -Parts:list, -Owners:list, -Adjectives:list) is det.
%
% @param Conditions
% @param Mains
% @param Parts
% @param Owners
% @param Adjectives
%
conditions_nountype([], [], [], [], []).

conditions_nountype([H | T], Mains, [Part | Parts], Owners, Adjectives) :-
	H = has_part(_, Part),
	!,
	conditions_nountype(T, Mains, Parts, Owners, Adjectives).

conditions_nountype([H | T], Mains, Parts, [H | Owners], Adjectives) :-
	functor(H, relation, _),
	!,
	conditions_nountype(T, Mains, Parts, Owners, Adjectives).

conditions_nountype([H | T], Mains, Parts, Owners, [H | Adjectives]) :-
	functor(H, Functor, _),
	(
		Functor = property
	;
		Functor = query
	),
	!,
	conditions_nountype(T, Mains, Parts, Owners, Adjectives).

% BUG: Here we convert those atoms that are in fact numbers (e.g. '5') into actual numbers (e.g. 5).
% APE does not generate such "atom numbers", but other DRS generators might.
conditions_nountype(
		[object(Referent, Value, Quantisation, Unit, Operator, Count) | T],
		[object(Referent, Value, Quantisation, Unit, Operator, NCount) | Mains], Parts, Owners, Adjectives
	) :-
	atom(Count), Count \= na,
	!,
	atom_number(Count, NCount),
	conditions_nountype(T, Mains, Parts, Owners, Adjectives).

% Fall back: other conditions go to Mains.
conditions_nountype([H | T], [H | Mains], Parts, Owners, Adjectives) :-
	conditions_nountype(T, Mains, Parts, Owners, Adjectives).


%% conditions_verbtype(+Conditions:list, -Predicates:list, -Adverbs:list, -PPs:list) is det.
%
% @param Conditions
% @param Predicates
% @param Adverbs
% @param PPs
%
% BUG: Note that we treat query/2 as a PP because it seems to act more like a PP, but
% this needs to be investigated further.
%
conditions_verbtype([], [], [], []).

conditions_verbtype([H | T], Predicates, [H | Adverbs], PPs) :-
	functor(H, modifier_adv, _),
	!,
	conditions_verbtype(T, Predicates, Adverbs, PPs).

conditions_verbtype([H | T], Predicates, Adverbs, [H | PPs]) :-
	functor(H, Functor, _),
	(
		Functor = modifier_pp
	;
		Functor = query
	),
	!,
	conditions_verbtype(T, Predicates, Adverbs, PPs).

% Fall back: other conditions go to Predicates.
conditions_verbtype([H | T], [H | Predicates], Adverbs, PPs) :-
	conditions_verbtype(T, Predicates, Adverbs, PPs).


%% make_boxid(-ID:integer) is det.
%
% @param ID is a DRS box identifier
%
% Generate IDs. If no ID has been generated yet then start with 0.
%
make_boxid(NewID) :-
	retract(boxid(OldID)),
	NewID is OldID + 1,
	assertz(boxid(NewID)),
	!.

make_boxid(0) :-
	assertz(boxid(0)).


%% make_ref2conds(+Drs:drs) is det.
%
% @param Drs is an Attempto DRS
%
% We assert all discourse referents from DRS into ref2conds/3 to speed up lookup later.
%
make_ref2conds(drs(Dom, Conds)) :-
	retractall(ref2conds(_, _, _, _)),
	retractall(boxid(_)),
	make_boxid(Id),
	assert_referents(Dom, Conds, [Id, top, []]),
	walk(Conds, [Id, top, []]).


%% assert_referents(+Referents:list, +ConditionList:list, +BoxId:term) is det.
%
% @param Referents are discourse referents of the DRS box
% @param ConditionList are conditions of the DRS box
% @param BoxId is a unique ID of the box
%
% Finds all the conditions in the ConditionList that have the same Referent
% as the first argument. Asserts the predicate =|ref2conds/4|= containing
% information about the shared Referent, the conditions that share the Referent,
% and the DRS-box ID.
% Finally processes separately the remaining conditions: in case the remaining
% conditions have the form formula/3 then they are asserted as they are.
% In case the remaining conditions are complex conditions then they are just ignored.
% Otherwise (e.g. in case of relation/3) a failure is triggered.
%
% BUG: maybe complex conditions should not reach this rule.
%
assert_referents([], Conditions, BoxId) :-
	assert_rest(Conditions, BoxId).

assert_referents([Referent | ReferentsTail], Conditions, BoxId) :-
	mypartition(Referent, Conditions, Included, Excluded, Id),
	assert(ref2conds(Referent, Included, BoxId, Id)),
	assert_referents(ReferentsTail, Excluded, BoxId).


%% mypartition(Referent, Conditions1, Conditions2, Excluded) is det.
%
% BUG: document
% replace by SWI-Prolog built-in partition/4 (available in recent versions)
%
mypartition(_Referent, [], [], [], _).

mypartition(Referent, [Condition-Id | Conditions], [Condition | Included], Excluded, Id) :-
	Condition \= formula(_, _, _),
	arg(1, Condition, Referent),
	!,
	mypartition(Referent, Conditions, Included, Excluded, Id).

mypartition(Referent, [Condition | Conditions], Included, [Condition | Excluded], Id) :-
	mypartition(Referent, Conditions, Included, Excluded, Id).


%% assert_rest(+CondList:list, +BoxId:term) is det.
%
%
assert_rest([], _).

assert_rest([formula(Expr1, Eq, Expr2)-Id | CondList], BoxId) :-
	assert(ref2conds(formula(Expr1, Eq, Expr2), [formula(Expr1, Eq, Expr2)], BoxId, Id)),
	assert_rest(CondList, BoxId).

assert_rest([relation(_, _, _)-_ | _CondList], _BoxId) :-
	!,
	fail.

assert_rest([has_part(_, _)-_ | _CondList], _BoxId) :-
	!,
	fail.

assert_rest([_Cond | CondList], BoxId) :-
	assert_rest(CondList, BoxId).


%% walk(+Conditions:list, +BoxId:term) is det.
%
% Traverses the DRS to assert all the referents together with their
% conditions. We also generate an ID for each embedded DRS (i.e. a box).
%
% @param Conditions are conditions of the DRS box
% @param BoxId is a unique ID of the box
%
% @bug Specify the BoxId better.
%
walk([], _).

walk([-drs(Dom, Conds1) | Conds2], BoxId) :-
	make_boxid(Id),
	assert_referents(Dom, Conds1, [Id, not, BoxId]),
	walk(Conds1, [Id, not, BoxId]),
	walk(Conds2, BoxId).

walk([Label:drs(Dom, Conds1) | Conds2], BoxId) :-
	assert(ref2conds(Label, [drs(Dom, Conds1)], BoxId, 'BUG')),
	walk(Conds2, BoxId).

walk([~drs(Dom, Conds1) | Conds2], BoxId) :-
	make_boxid(Id),
	assert_referents(Dom, Conds1, [Id, naf, BoxId]),
	walk(Conds1, [Id, naf, BoxId]),
	walk(Conds2, BoxId).

walk([<>(drs(Dom, Conds1)) | Conds2], BoxId) :-
	make_boxid(Id),
	assert_referents(Dom, Conds1, [Id, can, BoxId]),
	walk(Conds1, [Id, can, BoxId]),
	walk(Conds2, BoxId).

walk([[](drs(Dom, Conds1)) | Conds2], BoxId) :-
	make_boxid(Id),
	assert_referents(Dom, Conds1, [Id, must, BoxId]),
	walk(Conds1, [Id, must, BoxId]),
	walk(Conds2, BoxId).

walk([drs(Dom1, Conds1) => drs(Dom2, Conds2) | Conds3], BoxId) :-
	make_boxid(Id1),
	assert_referents(Dom1, Conds1, [Id1, if, BoxId]),
	walk(Conds1, [Id1, if, BoxId]),
	make_boxid(Id2),
	assert_referents(Dom2, Conds2, [Id2, then, BoxId]),
	walk(Conds2, [Id2, then, BoxId]),
	walk(Conds3, BoxId).

walk([drs(Dom1, Conds1) v drs(Dom2, Conds2) | Conds3], BoxId) :-
	make_boxid(Id1),
	assert_referents(Dom1, Conds1, [Id1, or1, BoxId]),
	walk(Conds1, [Id1, or1, BoxId]),
	make_boxid(Id2),
	assert_referents(Dom2, Conds2, [Id2, or2, BoxId]),
	walk(Conds2, [Id2, or2, BoxId]),
	walk(Conds3, BoxId).

walk([_-_ | Conds], BoxId) :-
	walk(Conds, BoxId).


%% get_box_prefix(BoxId:term, BoxPrefix:list) is det.
%
% @param BoxId is ...
% @param BoxPrefix is ...
%
% In case any Referent from a given box has already been
% verbalized then we conjoin the verbalization of the current Referent.
%
% In case no Referent from a given box has been verbalized,
% then the current Referent is the first one and we
% initiate the box. Then we check the outer box to come
% up for the name of the box (and do this recursively).


% Not the first Referent in the top-box to be verbalized.
get_box_prefix([0, top, []], ['.']) :-
	is_rep(0),
	!.

% First Referent in the top-box to be verbalized.
get_box_prefix([0, top, []], []) :-
	assert(is_rep(0)),
	!.

% Not the first Referent in the OR1-box to be verbalized.
% Given that the OR1 V OR2 is in the NOT-box.
get_box_prefix([Id, or1, [_, Marker, _]], [and, that]) :-
	is_unary_modifier(Marker),
	is_rep(Id),
	!.

% Not the first Referent in the NOT-box to be verbalized.
get_box_prefix([Id, Marker, _], [and, that]) :-
	is_unary_modifier(Marker),
	is_rep(Id),
	!.

% BUG: ???
% Not the first Referent in the any box to be verbalized.
get_box_prefix([Id, _, _], [and]) :-
	is_rep(Id),
	!.

% The first Referent in the THEN-box to be verbalized.
get_box_prefix([Id, then, _], [then]) :-
	assert(is_rep(Id)),
	!.

% The first Referent in the OR2-box to be verbalized.
% Given that the OR1 V OR2 is in the NOT-box.
% BUG: do we have to worry about any higher-level NOT-boxes
% or is checking just for the parent NOT-box OK?
get_box_prefix([Id, or2, [_, Marker, _]], [or, that]) :-
	is_unary_modifier(Marker),
	assert(is_rep(Id)),
	!.

% The first Referent in the OR1-box.
% The parent is the NOT-box.
% The parent has already a Referent.
get_box_prefix([Id, or1, [ParentId, Marker, _]], [',', and, that]) :-
	is_unary_modifier(Marker),
	is_rep(ParentId),
	assert(is_rep(Id)),
	!.

% The first Referent in the OR1-box.
% The parent is not the NOT-box and is not the top-box.
% The parent has already a Referent.
get_box_prefix([Id, or1, [ParentId, _, _]], [',', and]) :-
	ParentId \= 0,
	is_rep(ParentId),
	assert(is_rep(Id)),
	!.

% The first Referent in the OR2-box to be verbalized.
% Given that the parent is not a NOT-box.
get_box_prefix([Id, or2, _], [or]) :-
	assert(is_rep(Id)),
	!.

% The general case. The first Referent to be verbalized in either:
% IF-box, OR1-box, NOT-box.
get_box_prefix([Id, BoxMarker, OuterBoxId], [OuterBoxPrefix | BoxPrefix]) :-
	assert(is_rep(Id)),
	get_box_name(BoxMarker, BoxPrefix),
	get_box_prefix(OuterBoxId, OuterBoxPrefix).


%% is_unary_modifier(+Type:atom) is det.
%
% @param Type is in {not, naf, can, must}

is_unary_modifier(not).
is_unary_modifier(naf).
is_unary_modifier(can).
is_unary_modifier(must).


%% get_box_name(+BoxMarker:atom, -BoxPrefix:list) is det.
%
% @param BoxMarker is ...
% @param BoxPrefix is a list of ACE tokens associated with the box
%
% This is the conjunction with a preceding box.
%
get_box_name(if, [if]).
%%get_box_name(then, [then]).
get_box_name(or1, []).
%%get_box_name(or2, [or]).
get_box_name(not, [it, is, false, that]).
get_box_name(naf, [it, is, not, provable, that]).
get_box_name(can, [it, is, possible, that]).
get_box_name(must, [it, is, necessary, that]).


%% conds_text(+Conds:term, +BoxPrefix:atom, +TopDeep:atom, -Text:list, -Features:list) is det.
%
% Note: some discourse referents don't have surface representation.
% We don't call such referents at all.
%
% Note: some objects can have an owner, but in this case max 1 owner.
% ProperName objects cannot have an owner.
%
% @param Conds is a complex term with all the conditions grouped
% @param BoxPrefix is a prefix of tokens to be inserted in front of the verbalization (e.g. `and if it is false that')
% @param TopDeep is one of {top, deep}
% @param Text is an ACE text that corresponds to the discourse referent
% @param Features is in {f(sg, subj), f(sg, obj), f(pl, subj), f(pl, obj)}, for number and case sharing


% Old somebody and something are referred to by a variable only.
conds_text(noun([object(A, somebody, countable, na, eq, 1)], parts([]), owners([]), adjectives([])), _, deep, RI--RI, [A], f(sg, _)) :-
	memberchk(A, RI),
	add_var(A),
	!.

conds_text(noun([object(A, something, dom, na, na, na)], parts([]), owners([]), adjectives([])), _, deep, RI--RI, [A], f(sg, _)) :-
	memberchk(A, RI),
	add_var(A),
	!.


% New somebody and something (with 'there is')
conds_text(noun([object(A, somebody, countable, na, eq, 1)], parts([]), owners([]), adjectives([])),
BoxPrefix, top, RI--[A | RI], [BoxPrefix, there, is, somebody, A], f(sg, _)) :-
	\+ memberchk(A, RI),
	add_var(A),
	!.

conds_text(noun([object(A, something, dom, na, na, na)], parts([]), owners([]), adjectives([])),
BoxPrefix, top, RI--[A | RI], [BoxPrefix, there, is, something, A], f(sg, _)) :-
	\+ memberchk(A, RI),
	add_var(A),
	!.

% New somebody and something.
conds_text(noun([object(A, somebody, countable, na, eq, 1)], parts([]), owners([]), adjectives([])), _, deep, RI--[A | RI], [somebody, A], f(sg, _)) :-
	\+ memberchk(A, RI),
	add_var(A),
	!.

conds_text(noun([object(A, something, dom, na, na, na)], parts([]), owners([]), adjectives([])), _, deep, RI--[A | RI], [something, A], f(sg, _)) :-
	\+ memberchk(A, RI),
	add_var(A),
	!.


% NP conjunction. This is the only case where parts/1 has a non-empty list as an argument
% Old plural object, refer to it by 'they' (and 'them'). E.g.
% John and Mary wait. They talk.
% John and Mary wait. Bill sees them.
% * John and Mary wait. John and Mary talk. (this is not supported by refres)
%
% BUG: `them' is not supported
conds_text(noun([object(A, na, countable, _, _, _)], parts([_ | _]), _Owner, _Adjectives), _BoxPrefix, deep, RI--RI, [they], f(pl, subj)) :-
	memberchk(A, RI),
	!.

conds_text(noun([object(A, na, countable, _, _, _)], parts([_ | _]), _Owner, _Adjectives), _BoxPrefix, deep, RI--RI, [them], f(pl, obj)) :-
	memberchk(A, RI),
	!.

% NP conjunction.
% Note that we require that parts/1 has a non-empty list as an argument and that the number of
% arguments matches the number specified in the object-condition (QNum).
% In the DRSs generated by APE, this is not always the case. Consider DRSs where the
% has_part/2 condition is embedded into negation, such as the ones generated from:
% There are less than 3 men and less than 3 women.
% Less than 3 men and John wait.
conds_text(noun([object(A, na, countable, _, eq, QNum)], parts(ListOfObjects), _Owner, _Adjectives), BoxPrefix, TopDeep, RI--RO, [Prefix, ListOfObjectsText], f(pl, _)) :-
	!,
	debug(npcoord, 'npcoord: QNum: ~w; ListOfObjects: ~w~n', [QNum, ListOfObjects]),
	length(ListOfObjects, QNum),
	surface_det(A, RI, pl, TopDeep, BoxPrefix, countable, dummy, dummy, Prefix),
	objects_to_text(ListOfObjects, [A | RI]--RO, ListOfObjectsText),
	debug(verbose, 'top:~w prefix:~w result:~w~n', [TopDeep, BoxPrefix, Prefix]).

% Which men wait?
% There are which men of John?
conds_text(noun([object(A, Agent, countable, na, geq, 2)], parts([]), Owner, adjectives([query(A, which)])),
BoxPrefix, TopDeep, RI--RO, [ThereIs, qp(which), AgentText, OwnerText], f(pl, _)) :-
	!,
	\+ memberchk(A, RI),
	ref2conds(A, _, [0, top, []], _),
	get_thereis(TopDeep, pl, BoxPrefix, ThereIs),
	surface_noun(cn, pl, Agent, AgentText),
	verbalize_owners(A, RI--RO, Owner, OwnerText).

% BUG: temporary handling of "less than" and "at most"
conds_text(noun([object(_, _, countable, na, na, na)], parts([]), _, adjectives(_)), _, _, _, _, _) :-
	!,
	fail.

conds_text(noun([object(A, Agent, countable, na, Comp, Number)], parts([]), Owner, adjectives(Adjectives)),
	BoxPrefix, TopDeep, RI--RO, [SurfaceDeterminer, AdjectivesText, AgentText, A, OwnerText], f(pl, _)) :-
	Number > 1,
	surface_det(A, RI, pl, TopDeep, BoxPrefix, countable, Comp, Number, SurfaceDeterminer),
	surface_noun(cn, pl, Agent, AgentText),
	adjectives_to_text(Adjectives, AdjectivesText),
	verbalize_owners(A, RI--RO, Owner, OwnerText),
	add_var(A).


% Singular and mass nouns
%
% 1. Indefinite pronouns (somebody, something) can have owners but not properties.
%
% Everything of John is red.
% If there is something B of John then B is red.
%
% Note that this wouldn't work with quoted strings since they cannot have attached variables:
%
% Everything "Go!" of John is red. -> If there is something "Go!" of John then ??? is red.
%
% (Fortunately we don't support quoted strings in apposition anymore.)
%
% 2. Regular nouns
% 3. BUG: measurement nouns

% If there is somebody of John then ...
conds_text(noun([object(A, Value, _, _, _, _)], parts([]), Owner, _Adjectives), BoxPrefix, top,
	RI--RO, [BoxPrefix, there, is, Value, A, OwnerText], f(sg, _)) :-
	(Value = somebody ; Value = something),
	verbalize_owners(A, RI--RO, Owner, OwnerText),
	add_var(A).


% somebody of John waits ...
conds_text(noun([object(A, Value, _, _, _, _)], parts([]), Owner, _Adjectives), _BoxPrefix, deep,
	RI--RO, [Name, A, OwnerText], f(sg, _)) :-
	(Value = somebody ; Value = something),
	(
		memberchk(A, RI)
	->
		Name = []
	;
		Name = Value
	),
	verbalize_owners(A, RI--RO, Owner, OwnerText),
	add_var(A).


% Countable: which man?
% Which man waits?
% There is which man of John?
% (Note that `which' is not possible with mass nouns, qeneralized quantifiers, and indefinite pronouns.)
conds_text(noun([object(A, Agent, countable, na, eq, 1)], parts([]), Owner, adjectives([query(A, which)])),
BoxPrefix, TopDeep, RI--RO, [ThereIs, qp(which), AgentText, OwnerText], f(sg, _)) :-
	!,
	\+ memberchk(A, RI),
	ref2conds(A, _, [0, top, []], _),
	get_thereis(TopDeep, sg, BoxPrefix, ThereIs),
	surface_noun(cn, sg, Agent, AgentText),
	verbalize_owners(A, RI--RO, Owner, OwnerText).


% Countable: a man
conds_text(noun([object(A, Agent, countable, na, eq, 1)], parts([]), Owner, adjectives(Adjectives)),
	BoxPrefix, TopDeep, RI--RO, [SurfaceDeterminer, AdjectivesText, AgentText, A, OwnerText], f(sg, _)) :-
	surface_noun(cn, sg, Agent, AgentText),
	adjectives_to_text(Adjectives, AdjectivesText),
	surface_det(A, RI, sg, TopDeep, BoxPrefix, countable, _, _, SurfaceDeterminer),
	verbalize_owners(A, RI--RO, Owner, OwnerText),
	add_var(A).


% Countable: At least 1 man
conds_text(noun([object(A, Value, countable, na, Eq, 1)], parts([]), Owner, adjectives(Adjectives)),
	BoxPrefix, TopDeep, RI--RO, [SurfaceDeterminer, AdjectivesText, AgentText, A, OwnerText], f(sg, _)) :-
	Eq \= eq,
	surface_noun(cn, sg, Value, AgentText),
	adjectives_to_text(Adjectives, AdjectivesText),
	surface_det(A, RI, sg, TopDeep, BoxPrefix, countable, Eq, 1, SurfaceDeterminer),
	verbalize_owners(A, RI--RO, Owner, OwnerText),
	add_var(A).


% Mass: some rice
conds_text(noun([object(A, Agent, mass, na, na, na)], parts([]), Owner, adjectives(Adjectives)),
BoxPrefix, TopDeep, RI--RO, [SurfaceDeterminer, AdjectivesText, AgentText, A, OwnerText], f(sg, _)) :-
	surface_noun(cn, sg, Agent, AgentText),
	adjectives_to_text(Adjectives, AdjectivesText),
	surface_det(A, RI, sg, TopDeep, BoxPrefix, mass, _, _, SurfaceDeterminer),
	verbalize_owners(A, RI--RO, Owner, OwnerText),
	add_var(A).

% Measurement constructs, e.g.
% More than 3 kg of white rice rots.
% 2 kg of water boils. The water is warm.
% 2 kg of apples are ripe. The apples rot.
conds_text(noun([object(A, Agent, Quant, Unit, QType, QNum)], parts([]), Owner, adjectives(Adjectives)),
BoxPrefix, TopDeep, RI--RO, [ThereIs, Specifier, AdjectivesText, AgentText, A, OwnerText], f(SgPl, _)) :-
	(Unit \= na),
	(
		Quant = countable
	->
		SgPl = pl
	;
		SgPl = sg
	),
	(
		memberchk(A, RI)
	->
		ThereIs = [],
		surface_determiner(old, _, _, Specifier),
		OwnerText = []
	;
		get_thereis(TopDeep, SgPl, BoxPrefix, ThereIs),
		surface_determiner(new, QType, QNum, Q),
		Specifier = [Q, Unit, of],
		verbalize_owners(A, RI--RO, Owner, OwnerText)
	),
	adjectives_to_text(Adjectives, AdjectivesText),
	surface_noun(cn, SgPl, Agent, AgentText),
	add_var(A).


% Singular proper name
% Example: John waits.
% Note: ACE does not support proper name "owners" (as in "John of a cat waits.")?
conds_text(noun([object(_Referent, Value, named, _, _, 1)],
	parts([]), owners([]), adjectives([])), _, deep,
	RI--RI, [ValueText], f(sg, _)) :-
	surface_noun(pn, sg, Value, ValueText).

% Plural proper name
% Example: Christmases wait.
conds_text(noun([object(_Referent, Value, named, _, _, Number)],
	parts([]), owners([]), adjectives([])), _, deep,
	RI--RI, [ValueText], f(pl, _)) :-
	Number > 1,
	surface_noun(pn, pl, Value, ValueText).


% Intransitive verbs with (adverb | pp)* attachment.
conds_text(verb([predicate(_, Predicate, Argument)], adverbs(Adverbs), pps(PPs)), Box, _,
	RI--RO, [Box, ArgumentText, PredicateText, AdverbsText, PPsText], _) :-
	referent_text(Argument, RI--RTmp, ArgumentText, f(SgPl, subj)),
	surface_verb(SgPl, Predicate, PredicateText),
	adverbs_to_text(Adverbs, AdverbsText),
	pps_to_text(PPs, RTmp--RO, PPsText).

% Transitive verbs with (adverb | pp)* attachment.
% BUG: Comment is currently not returned.
conds_text(verb([predicate(_, Predicate, Argument1, Argument2)], adverbs(Adverbs), pps(PPs)), Box, _,
	RI--RO, [Box, Argument1Text, PredicateText, Argument2Text, AdverbsText, PPsText], _) :-
	debug(verbose, "transitive verb: referents in: ~w~n", [RI]),
	referent_text(Argument1, RI--RTmp1, Argument1Text, f(SgPl, subj)),
	surface_verb(SgPl, Predicate, PredicateText),

	referent_text(Argument2, RTmp1--RTmp2, Argument2Text, f(_, obj)),

	adverbs_to_text(Adverbs, AdverbsText),
	pps_to_text(PPs, RTmp2--RO, PPsText),
	make_comment(PredicateText, PPsText, _Comment).

% Ditransitive verbs with (adverb | pp)* attachment.
% Note: the order of calling referent_text/3 is important as assert/1 is performed by this call.
% Test it with "A man gives a dog the dog.", in the paraphrase the order of indef-NP and def-NP must be correct.
%
% BUG: Comment is currently not returned.
conds_text(verb([predicate(_, Predicate, Argument1, Argument2, Argument3)], adverbs(Adverbs), pps(PPs)), Box, _,
	RI--RO, [Box, Argument1Text, PredicateText, Argument3Text, Argument2Text, AdverbsText, PPsText], _) :-
	referent_text(Argument1, RI--RTmp1, Argument1Text, f(SgPl, subj)),
	surface_verb(SgPl, Predicate, PredicateText),

	referent_text(Argument3, RTmp1--RTmp2, Argument3Text, f(_, obj)),
	referent_text(Argument2, RTmp2--RTmp3, Argument2Text, f(_, obj)),

	adverbs_to_text(Adverbs, AdverbsText),
	pps_to_text(PPs, RTmp3--RO, PPsText),
	make_comment(PredicateText, PPsText, _Comment).


% Properties as copula arguments.
% Must have at least 1 element.
conds_text([FirstProperty | RestProperties], _, _, RI--RO, [PropertiesText], _) :-
	properties_to_text([FirstProperty | RestProperties], RI--RO, PropertiesText).



%% verbalize_owners(+Referent:nvar, +Owners:term, -TokenList:list) is det.
%
% @param Referent is a discourse referent
% @param Owners is a term =|owners([relation(_, of, OwnerReferent)])|=.
% Note that is always contains just one relation/3 condition.
% @param TokenList is a list of resulting ACE tokens
%
% Note that we verbalize the of-constructions only in case the
% NP is not anaphoric, i.e. we never output "the dog X of the man Y"
% but simply "the dog X".
%
verbalize_owners(Ref, RI--RI, owners([]), []) :-
	memberchk(Ref, RI),
	!.

verbalize_owners(Ref, RI--[Ref | RI], owners([]), []) :-
	!.

verbalize_owners(Ref, RI--RI, owners([_]), []) :-
	memberchk(Ref, RI),
	!.

verbalize_owners(Ref, RI--RO, owners([relation(_, of, OwnerReferent)]), [of, OwnerText]) :-
	referent_text(OwnerReferent, [Ref | RI]--RO, OwnerText, _Features).


%% adjectives_to_text(+PropertyList:list, -TokenList:list) is det.
%
% Adjectives are conjoined.
%
adjectives_to_text([], []).

adjectives_to_text([property(_, Property, Comparison)], PropertyText) :-
	!,
	surface_property(Property, Comparison, PropertyText).

adjectives_to_text([property(_, Property, Comparison) | Adjectives], [PropertyText, and | AdjectivesText]) :-
	surface_property(Property, Comparison, PropertyText),
	adjectives_to_text(Adjectives, AdjectivesText).


%% adverbs_to_text(+AdverbList:list, -AdverbListText:list) is det.
%
% @param AdverbList is a list of modifier_adv/3 conditions.
% @param AdverbListText is list of ACE adverbs conjoined by 'and'.
%
adverbs_to_text([], []).

adverbs_to_text([modifier_adv(_, Adverb, Comparison)], [AdverbText]) :-
	!,
	surface_adverb(Adverb, Comparison, AdverbText).

adverbs_to_text([modifier_adv(_, Adverb, Comparison) | AdverbList], [AdverbText, 'and' | AdverbListText]) :-
	surface_adverb(Adverb, Comparison, AdverbText),
	adverbs_to_text(AdverbList, AdverbListText).


%% pps_to_text(+Modifiers:list, -TokenList:list) is det.
%
% @param Modifiers is a list of modifier_pp- or query-conditions
% @param TokenList is a list of ACE tokens
%
% Note that we only support top-level ([0, top, []]) query-conditions ('how', 'where', 'when').
%
pps_to_text([], RI--RI, []).

pps_to_text([modifier_pp(_, Prep, Modifier) | PPs], RI--RO, [Prep, ModifierText | PPsText]) :-
	referent_text(Modifier, RI--RTmp, ModifierText, _Features),
	pps_to_text(PPs, RTmp--RO, PPsText).

pps_to_text([query(Referent, QueryPronoun) | PPs], RI--RO, [qp(QueryPronoun) | PPsText]) :-
	ref2conds(Referent, _, [0, top, []], _),
	pps_to_text(PPs, RI--RO, PPsText).


%% properties_to_text(+Properties:list, -TokenList:list) is det.
%
% @param Properties is a list of property-conditions
% @param TokenList is a list of ACE tokens
%
properties_to_text([], RI--RI, []).

properties_to_text([property(_, Property, Comparison)], RI--RI, PropertyText) :-
	!,
	surface_property(Property, Comparison, PropertyText).

properties_to_text([property(_, Property, Comparison, Argument)], RI--RO, [PropertyText, ArgumentText]) :-
	!,
	surface_property(Property, Comparison, PropertyText),
	referent_text(Argument, RI--RO, ArgumentText, _Features).

properties_to_text([property(_Ref1, Adjective, Ref2, Comparison, ComparisonTarget, Ref3)], RI--RO, PropertyText) :-
	!,
	referent_text(Ref2, RI--RTmp, Ref2Text, _),
	referent_text(Ref3, RTmp--RO, Ref3Text, _),
	surface_property(Adjective, Comparison, ComparisonTarget, Ref2Text, Ref3Text, PropertyText).

properties_to_text([property(_, Property, Comparison) | Properties], RI--RO, [PropertyText, 'and' | PropertiesText]) :-
	!,
	surface_property(Property, Comparison, PropertyText),
	properties_to_text(Properties, RI--RO, PropertiesText).


%% objects_to_text(+ListOfObjects:list, -ListOfObjectsVerbalized:list) is det.
%
% @param ListOfObjects is a list of variables and/or data items
% @param ListOfObjectsVerbalized is a list of ACE tokens
%
% Handling of plural objects which are built from several NPs.
%
objects_to_text([Agent], RI--RO, [AgentText]) :-
	referent_text(Agent, RI--RO, AgentText, _Features).

objects_to_text([Agent | Agents], RI--RO, [AgentText, 'and' | AgentsText]) :-
	referent_text(Agent, RI--RTmp, AgentText, _Features),
	objects_to_text(Agents, RTmp--RO, AgentsText).


%% surface_det(+Referent:term, +SgPl:atom, +TopDeep:atom, +BoxPrefix:list, +NounType:atom, +Comp:atom, +Number:atom, -Result:list) is det.
%
% @param Referent is a discourse referent
% @param SgPl is one of {sg, pl}
% @param TopDeep {top, deep, BUG}
% @param BoxPrefix BUG
% @param NounType {mass, countable}
% @param Comp {eq, geq, leq, greater, less}
% @param Number is an integer (larger than 0, but we are not checking this...)
% @param Result is a list of ACE tokens (can also be an atom)
%
% BUG: "top" means always "new"?

surface_det(Referent, RI, SgPl, TopDeep, BoxPrefix, NounType, Comp, Number, Result) :-
	(
		memberchk(Referent, RI)
	->
		OldNew = old
	;
		OldNew = new
	),
	surface_det_x(SgPl, TopDeep, BoxPrefix, NounType, OldNew, Comp, Number, Result).


surface_det_x(sg, top, BoxPrefix, countable, new, eq, 1, [BoxPrefix, there, is, a]).

surface_det_x(sg, top, BoxPrefix, mass, new, na, na, [BoxPrefix, there, is, some]).

surface_det_x(sg, deep, _, countable, new, eq, 1, [a]) :- !.
surface_det_x(sg, deep, _, mass, new, na, na, [some]).
surface_det_x(sg, deep, _, mass, old, _, _, [the]).


% There are John and Mary.
surface_det_x(pl, top, BoxPrefix, _, _, dummy, dummy, [BoxPrefix, there, are]) :- !.

% John and Mary wait.
surface_det_x(pl, deep, _, _, _, dummy, dummy, []) :- !.


surface_det_x(sg, top, BoxPrefix, countable, Det, Comp, Number, [BoxPrefix, there, is, SurfaceDeterminer]) :-
	surface_determiner(Det, Comp, Number, SurfaceDeterminer).

surface_det_x(sg, deep, _, countable, Det, Comp, Number, SurfaceDeterminer) :-
	surface_determiner(Det, Comp, Number, SurfaceDeterminer).


surface_det_x(pl, top, BoxPrefix, _, Det, Comp, Number, [BoxPrefix, there, are, SurfaceDeterminer]) :-
	surface_determiner(Det, Comp, Number, SurfaceDeterminer).

surface_det_x(pl, deep, _, _, Det, Comp, Number, SurfaceDeterminer) :-
	surface_determiner(Det, Comp, Number, SurfaceDeterminer).


%% surface_determiner(+OldNew:atom, +Operator:atom, +Count:atomic, -SurfaceDeterminer:list) is det.
%
% @param OldNew
% @param Operator is in {eq, geq, leq, greater, less, exactly, na}
% @param Count is a positive integer or 'na'
% @param SurfaceDeterminer is a list of tokens corresponding the an ACE determiner 
%
surface_determiner(old, _, _, the).

surface_determiner(new, eq, Number, [Number]).

surface_determiner(new, geq, Number, [at, least, Number]).

surface_determiner(new, leq, Number, [at, most, Number]).

surface_determiner(new, greater, Number, [more, than, Number]).

surface_determiner(new, less, Number, [less, than, Number]).

surface_determiner(new, exactly, Number, [exactly, Number]).

surface_determiner(new, na, na, []).


%% get_thereis(+TopDeep:atom, +SgPl:atom, +BoxPrefix:list, -ThereIs:list) is det.
%
% @param TopDeep is one of {top, deep}
% @param SgPl is one of {sg, pl}
% @param BoxPrefix
% @param ThereIs is a list of tokens containing the BoxPrefix and 'there is/are' is needed
%
get_thereis(top, sg, BoxPrefix, [BoxPrefix, there, is]).
get_thereis(top, pl, BoxPrefix, [BoxPrefix, there, are]).
get_thereis(deep, _, _, []).


%% make_comment(+VerbText:list, +PpText:list, -Comment:list) is det.
%
% On the basis of the PP construct a comment for novices.
%
% BUG: Experimental.
%
make_comment(_, [], []).
make_comment(PredicateText, [H | Tail], ['/*', 'Did you know that in ACE prepositional phrases always modify the verb? I.e.', PredicateText, [H | Tail], '*/']).


%% expr_exprpp(+Expression:term, -ExpressionText:list) is det.
%
% @param Expression
% @param ExpressionText
%
expr_exprpp(string(String), RI--RI, StringText) :-
	!,
	atomic(String), % the content of string/1 is either an atom or a number
	surface_quotedstring(String, StringText).

expr_exprpp(int(Number), RI--RI, NumberText) :-
	!,
	integer(Number),
	with_output_to(atom(NumberText), format("~w", [Number])).

expr_exprpp(real(Number), RI--RI, NumberText) :-
	!,
	float(Number),
	with_output_to(atom(NumberText), format("~w", [Number])).

expr_exprpp(int(Number, Unit), RI--RI, [NumberText, Unit]) :-
	!,
	integer(Number),
	with_output_to(atom(NumberText), format("~w", [Number])).

expr_exprpp(real(Number, Unit), RI--RI, [NumberText, Unit]) :-
	!,
	float(Number),
	with_output_to(atom(NumberText), format("~w", [Number])).

% Note: it's ok to call findall where RI--RI because sets cannot introduce new referents
expr_exprpp(set(Set), RI--RI, ['{', CommaSetText, '}']) :-
	!,
	findall(ElText, (member(El, Set), expr_exprpp(El, RI--RI, ElText)), SetText),
	list_commalist(SetText, CommaSetText).

% Note: it's ok to call findall where RI--RI because lists cannot introduce new referents
expr_exprpp(list(List), RI--RI, ['[', CommaListText, ']']) :-
	!,
	findall(ElText, (member(El, List), expr_exprpp(El, RI--RI, ElText)), ListText),
	list_commalist(ListText, CommaListText).

expr_exprpp(expr(Op, Expr1, Expr2), RI--RO, ['(', Expr1Text, Op, Expr2Text, ')']) :-
	!,
	expr_exprpp(Expr1, RI--RTmp, Expr1Text),
	expr_exprpp(Expr2, RTmp--RO, Expr2Text).

expr_exprpp(Referent, RI--RO, Variable) :-
	get_type(Referent, Type, BoxId),
	get_box_prefix(BoxId, BoxPrefix),
	conds_text(Type, BoxPrefix, deep, RI--RO, Text, _),
	flatten(Text, FlatText),
	last(FlatText, Variable),
	debug(verbose, "NP in expr: ~w ~w~n", [Text, Variable]).


%% list_commalist(?List:list, ?CommaList:list) is nondet.
%
% @param List is a list of elements
% @param CommaList is a list where the elements are separated by explicit commas
%
list_commalist([], []).
list_commalist([X], [X]).
list_commalist([X1, X2 | Tail], [X1, ',' | Text]) :-
	list_commalist([X2 | Tail], Text).


%% sentencelist_sentencelistcoord(+SentenceList:list, -SentenceListCoord:list) is det.
%
% Note that sentencelist cannot be empty.
%
sentencelist_sentencelistcoord([Sentence], [Sentence]).
sentencelist_sentencelistcoord([Sentence1, Sentence2 | SentenceList], [Sentence1, [and, that] | SentenceListCoord]) :-
	sentencelist_sentencelistcoord([Sentence2 | SentenceList], SentenceListCoord).


%% t(+AceText:atom) is det.
%% t(+AceText:atom, -Paraphrase:list) is det.
%
% @param AceText is an ACE text (as an atom)
% @param Paraphrase is a paraphrase of ACE text (as a list of sentences)
%
% Just for testing purposes.
%
t(Text) :-
	t(Text, Paraphrase),
	format("~w~n", [Paraphrase]).

t(Text, Paraphrase) :-
	ace_to_drs(Text, D, _M),
	typeddrs_to_untypeddrs(D, UntypedD),
	drs_to_coreace:drs_to_coreace(UntypedD, Paraphrase).
