% This file is part of the Attempto Parsing Engine (APE).
% Copyright 2008, Kaarel Kaljurand <kaljurand@gmail.com>.
%
% The Attempto Parsing Engine (APE) is free software: you can redistribute it and/or modify it
% under the terms of the GNU Lesser General Public License as published by the Free Software
% Foundation, either version 3 of the License, or (at your option) any later version.
%
% The Attempto Parsing Engine (APE) is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
% PURPOSE. See the GNU Lesser General Public License for more details.
%
% You should have received a copy of the GNU Lesser General Public License along with the Attempto
% Parsing Engine (APE). If not, see http://www.gnu.org/licenses/.


:- module(drs_to_owlswrl, [
		drs_to_owlswrl/2,
		drs_to_owlswrl/4,
		drs_to_owlswrl/5,
		condlist_and/4,
		is_toplevel/3,
		get_namedclass/2,
		get_namedobjectproperty/2,
		get_namedobjectproperty/3,
		get_nameddataproperty/2,
		get_nameddataproperty/3,
		is_object_with_generalized_quantifier/1
	]).


/** <module> Attempto DRS to OWL 2/SWRL translator

Translate an Attempto DRS into Web Ontology Language (OWL 2),
or if this fails then to Semantic Web Rule Language (SWRL).

If the translation fails then we search for errors by traversing
the respective structure (e.g. implication) again. Note that the error
capture is not completely implemented. Sometimes the translation simply
fails and no explanatory messages are asserted.

@author Kaarel Kaljurand
@version 2008-05-04
@license LGPLv3

TODO:

==
- general:
	- put "tricks" into a separate module
	- low priority: add a trick: if a DRS maps to a SWRL rule which uses sameAs/differentFrom,
	but replacing these with normal object properties would make the rule expressible
	directly in OWL (probably SubPropertyOf with property composition),
	then replace these with ace:equal/ace:differ-from (with equivalent 
	definitions) and express the rule in OWL syntax.

- allow:
	- URGENT: What is a country that borders no sea? (= What is a country that does not border a sea?)
	- Hungary is a country that borders no sea. (have a nicer output with less anon inds)
	- Every man is a human and sees the human.
	- John's age is not 21.
	- John's age is 21 or is 22.
	- Every man owns something that is "abc". (because we do allow: Every man owns "abc".)
	- For everything X for every thing Y if X likes Y then X does not hate Y.
	- Every dog hates every cat. (MAYBE)
	- Mary eats some grass. (because we do allow: Everybody that eats some meat is a carnivore.)
	- Every lady's pets are nothing but cats. (instead to/in addition to: Every lady's pet is nothing but cats.)
	- Every lady is somebody whose pets are more than 3 cats. (instead/in addition to: Every lady is somebody whose pet is more than 3 cats.)
	- John likes more than 3 women that own a car.
	- Every man is himself. (currently generates SWRL, but we could generate nothing in this case)

- check if the syntax is according to the spec:
	- DataProperty, DataValue, DataType

- reject:
	- DONE? If 2 men own a car then the men like Mary.
	(the object selected in the if-part should be (eq, 1)

- better error messages for:
	- No card is valid.
	- Every man does not have to cook a chicken.
	- Every singular is every singular. (from the log)
	- top-level negation
	- top-level disjunction

- improve:
	- implement namespaces support, i.e. each name should actually be a term ':'(NS, Name)
==
*/

:- use_module('../drs_to_sdrs', [
		drs_to_sdrs/2
	]).

:- use_module('../../logger/error_logger', [
		add_error_message/4,
		clear_messages/1
	]).

:- use_module(drs_to_owldrs, [
		drs_to_owldrs/2
	]).

:- use_module(implication_to_swrl, [
		implication_to_swrl/3
	]).

:- use_module(drs_to_dlquery, [
		condlist_to_dlquery/3
	]).

:- use_module(illegal_conditions, [
		illegal_condition/1,
		illegal_conditions/1
	]).

:- use_module(transform_anonymous, [
		transform_anonymous/2
	]).


%:- debug(owldrs).
%:- debug(sentence).


% Operators used in the DRS.
:- op(400, fx, -).
:- op(500, xfx, =>).
:- op(500, xfx, v).


%% drs_to_owlswrl(+Drs:term, -Owl:term) is semidet.
%% drs_to_owlswrl(+Drs:term, +Name:atom, -Owl:term) is semidet.
%% drs_to_owlswrl(+Drs:term, +Name:atom, +Comment:atom, -Owl:term) is semidet.
%% drs_to_owlswrl(+Drs:term, +Name:atom, +Comment:atom, -OwlDrs:term, -Owl:term) is semidet.
%
% Converts an Attempto DRS into OWL 2/SWRL.
% In the beginning, the DRS is modified in order to make
% the processing more straight-forward. The modified
% DRS (generated by drs_to_owldrs/2) is also returned (by drs_to_swrl/5).
%
% @param Drs is an Attempto DRS (untyped)
% @param Name is a URI relative to which all class names are to be interpreted
% @param Comment is a comment to be inserted into the resulting ontology (currently ignored)
% @param OwlDrs is a modified DRS, on the basis of which the OWL axioms are generated
% @param Owl is an OWL ontology in the form 'Ontology'(Name, AxiomList) where AxiomList
% is a list of axioms that correspond to the DRS conditions, in OWL FSS (Prolog notation)

drs_to_owlswrl(DrsIn, Owl) :-
	ontology_url(UrlPrefix),
	drs_to_owlswrl(DrsIn, UrlPrefix, Owl).

drs_to_owlswrl(DrsIn, Name, Owl) :-
	drs_to_owlswrl(DrsIn, Name, 'Ontology from an ACE text.', Owl).

drs_to_owlswrl(DrsIn, Name, Comment, Owl) :-
	drs_to_owlswrl(DrsIn, Name, Comment, _, Owl).

drs_to_owlswrl(DrsIn, Name, Comment, Drs, 'Ontology'(Name, AxiomList)) :-
	copy_term(DrsIn, DrsCopy),

	clear_messages(owl),
	debug(sentence, "ACE: ~w~n", [Comment]),
	drs_to_sdrs(DrsCopy, SDrsCopy),
	numbervars(SDrsCopy, 0, _),
	drs_to_owldrs(SDrsCopy, Drs),
	debug(owldrs, "OWL DRS: ~w~n", [Drs]),
	findall(
		ref_oneof(X, OneOf),
		(
			member(Condition, Drs),
			condition_oneof(Condition, X, OneOf)
		),
		RefList
	),
	debug(sentence, "Toplevel: ~w~n", [RefList]),
	condlist_axiomlist_with_cheat(Drs, RefList, AxiomListTmp),
	(
		memberchk(query(_, QueryWord)-Id, Drs)
	->
		clear_messages(owl),
		(
			condlist_to_dlquery(Drs, RefList, Class)
		->
			AL = ['SubClassOf'(Class, 'OWLClass'(owl:'Thing'))]
		;
			AL = [],
			add_error_message(owl, Id, QueryWord, 'Queries not supported.')
		)
	;
		AL = AxiomListTmp
	),
	transform_anonymous(AL, AxiomList),
	!.

drs_to_owlswrl(_, _, _, _, 'Ontology'('', [])).


%% condition_oneof(+Condition:term, -Ref:term, -ObjectOneOf:term) is semidet.
%
% Note that this rule is only called for the top-level DRS conditions.
%
% @param Condition is a DRS (top-level) condition
% @param Ref is a discourse referent
% @param ObjectOneOf is OWL's ObjectOneOf construct that corresponds to the condition
%
% @bug Fix comments, we also return genquant/4 now
%
condition_oneof(object(X, Name, countable, _, QType, QNum)-_, X, genquant(X, Name, QType, QNum)) :-
	is_object_with_generalized_quantifier(object(_, Name, countable, _, QType, QNum)-_),
	!.

condition_oneof(object(X, Name, named, _, _, _)-_, X, 'ObjectOneOf'([Individual])) :-
	get_namedindividual(Name, Individual).

condition_oneof(object(X, Name, countable, _, _, _)-_, X, 'ObjectOneOf'(['AnonymousIndividual'(X)])) :-
	Name \= na.

condition_oneof(object(X, something, dom, _, _, _)-_, X, 'ObjectOneOf'(['AnonymousIndividual'(X)])).


%% is_object_with_generalized_quantifier(+Condition:term) is semidet.
%
% Succeeds if the condition is an object-condition with certain QType and QNum arguments.
%
% Note: "Name \= na" is there to reject NP conjunctions.
%
% @param Condition is a DRS condition
%
is_object_with_generalized_quantifier(object(_, Name, countable, _, QType, QNum)-_) :-
	Name \= na,
	(
		QNum > 1
	;
		QNum = 1, QType = exactly
	;
		QNum = 1, QType = leq
	;
		QNum = 1, QType = less
	;
		QNum = 1, QType = greater
	).


%% condlist_axiomlist_with_cheat(+ConditionList:list, +RefList:list, -AxiomList:list) is semidet.
%
% Mapping from the DRS conditions' list to OWL axioms.
%
% @bug We first cheat a bit, e.g. in case the DRS consists of
% certain 3 conditions then they are mapped immediately to a
% class assertion. This avoids the anonymous individuals
% that are generated by the general solution.
%
% @bug Also, sentences like "There are exactly 4 continents." are handled here.
%
% @param ConditionList is a list of DRS conditions
% @param RefList is a list of top-level discourse referents
% @param AxiomList is a list of OWL axioms
%
condlist_axiomlist_with_cheat(
	[object(X, ProperName, named, _, eq, 1)-_, object(Y, Noun, NotNamed, _, _, _)-_, predicate(_, be, X, Y)-_],
	_,
	['ClassAssertion'(Class, Individual)]
) :-
	NotNamed \= named,
	get_namedclass(Noun, Class),
	get_namedindividual(ProperName, Individual),
	!.

condlist_axiomlist_with_cheat(
	[object(X, ProperName, named, _, eq, 1)-_, object(Y, Noun, NotNamed, _, QType, QNum)-_, predicate(_, ObjectPropertyName, X, Y)-_],
	_,
	['ClassAssertion'(PropertyRestriction, Individual)]
) :-
	ObjectPropertyName \= be,
	NotNamed \= named,
	get_namedclass(Noun, Class),
	get_namedobjectproperty(ObjectPropertyName, ObjectProperty),
	get_namedindividual(ProperName, Individual),
	make_restr(QType, QNum, ObjectProperty, Class, PropertyRestriction),
	!.

condlist_axiomlist_with_cheat(
	[object(X, ProperName, named, _, eq, 1)-_, object(Y, Noun, NotNamed, _, QType, QNum)-_, predicate(_, ObjectPropertyName, Y, X)-_],
	_,
	['ClassAssertion'(PropertyRestriction, Individual)]
) :-
	ObjectPropertyName \= be,
	NotNamed \= named,
	get_namedclass(Noun, Class),
	get_namedobjectproperty(ObjectPropertyName, ObjectProperty),
	get_namedindividual(ProperName, Individual),
	make_restr(QType, QNum, 'InverseObjectProperty'(ObjectProperty), Class, PropertyRestriction),
	!.

% There are exactly 4 continents.
% <= Everything is known by Superman. Superman knows exactly 4 continents.
% @bug: 'Superman' should be in a different namespace.
% @bug: 'know' should be in a different namespace.
% @bug: we shouldn't repeat the SubClassOf axiom, i.e. we could expect the client to contain it already.
condlist_axiomlist_with_cheat(
	[object(X, Noun, NotNamed, _, QType, QNum)-_],
	_,
	[
		'SubClassOf'(
			'OWLClass'(owl:'Thing'),
			'ObjectSomeValuesFrom'(
				'InverseObjectProperty'(ObjectProperty),
				'ObjectOneOf'([Individual])
			)
		),
		'ClassAssertion'(PropertyRestriction, Individual)
	]
) :-
	is_object_with_generalized_quantifier(object(X, Noun, NotNamed, _, QType, QNum)-_),
	get_namedclass(Noun, Class),
	get_namedobjectproperty(know, ace, ObjectProperty),
	get_namedindividual('Superman', ace, Individual),
	make_restr(QType, QNum, ObjectProperty, Class, PropertyRestriction),
	!.

% It is false that there are exactly 4 continents.
condlist_axiomlist_with_cheat(
	[-[object(X, Noun, NotNamed, _, QType, QNum)-_]],
	_,
	[
		'SubClassOf'(
			'OWLClass'(owl:'Thing'),
			'ObjectSomeValuesFrom'(
				'InverseObjectProperty'(ObjectProperty),
				'ObjectOneOf'([Individual])
			)
		),
		'ClassAssertion'('ObjectComplementOf'(PropertyRestriction), Individual)
	]
) :-
	is_object_with_generalized_quantifier(object(X, Noun, NotNamed, _, QType, QNum)-_),
	get_namedclass(Noun, Class),
	get_namedobjectproperty(know, ace, ObjectProperty),
	get_namedindividual('Superman', ace, Individual),
	make_restr(QType, QNum, ObjectProperty, Class, PropertyRestriction),
	!.

% @bug Cheating ends here, we call condlist_axiomlist/3
condlist_axiomlist_with_cheat(ConditionList, RefList, AxiomList) :-
	condlist_axiomlist(ConditionList, RefList, AxiomList).


%% condlist_axiomlist(+ConditionList:list, +RefList:list, -AxiomList:list) is semidet.
%
% @param ConditionList is a list of DRS conditions
% @param RefList is a list of top-level discourse referents
% @param AxiomList is a list of OWL axioms
%
% Mapping from the DRS conditions' list to OWL axioms.

% ConditionList is empty
condlist_axiomlist([], _, []).

% Condition is successfully mapped to an Axiom
condlist_axiomlist([Condition | ConditionList], RefList, [Axiom | AxiomList]) :-
	condition_axiom(Condition, RefList, Axiom),
	!,
	condlist_axiomlist(ConditionList, RefList, AxiomList).

% Condition is successfully mapped to a SWRL rule
condlist_axiomlist([Condition | ConditionList], RefList, [Axiom | AxiomList]) :-
	implication_to_swrl(Condition, RefList, Axiom),
	!,
	condlist_axiomlist(ConditionList, RefList, AxiomList).

% BUG: experimental: ignore objects that have a generalized quantifier
condlist_axiomlist([Condition | ConditionList], RefList, AxiomList) :-
	is_object_with_generalized_quantifier(Condition),
	!,
	condlist_axiomlist(ConditionList, RefList, AxiomList).

% A top-level condition that is not supported (e.g. has_part/2)
% is simply ignored, but an error message is asserted.
condlist_axiomlist([Condition | ConditionList], RefList, AxiomList) :-
	illegal_condition(Condition),
	!,
	condlist_axiomlist(ConditionList, RefList, AxiomList).

% If mapping of Condition failed and if the condition
% was an if-then condition then we search for the exact location
% of the error.
% BUG: we should do the same for negation and disjunction
condlist_axiomlist([If => Then | CondList], RefList, AxiomList) :-
	select(object(X, _Value, _, _, _, _)-_, If, CondList1),
	condlist_classlist_err(X, CondList1, RefList),
	condlist_classlist_err(X, Then, RefList),
	!,
	add_error_message(owl, '', '', 'Text could not be translated.'),
	condlist_axiomlist(CondList, RefList, AxiomList).

% If everything fails then set a general error message.
% BUG: Figure out the sentence ID at least.
condlist_axiomlist([_Condition | CondList], RefList, AxiomList) :-
	add_error_message(owl, '', '', 'Text could not be translated.'),
	condlist_axiomlist(CondList, RefList, AxiomList).


%% condition_axiom(+Condition:term, +RefList:list, -Axiom:term) is semidet.
%
% @param Condition is a DRS condition
% @param RefList is a list of top-level discourse referents
% @param Axiom is an OWL axiom
%
% Mapping the DRS conditions to OWL axioms. Note that each top-level
% condition maps to exactly one OWL axiom. This holds for the modified
% DRS where e.g. the relation-conditions have been removed and the implication
% corresponding
% to sentences like "If somebody X writes something Y then X is a writer and
% Y is a writing." have been split into 2 implications, correspoding to:
% Everybody who writes something is a writer." and "Everything that somebody
% writes is a writing."
%

% Negation.
% E.g. John is not Mary., John does not like a man who owns at most 2 cars.
condition_axiom(
	-Not,
	RefList,
	'ClassAssertion'('ObjectComplementOf'(NotClass), Individual)
) :-
	condlist_and(D, Not, RefList, NotClass),
	is_toplevel(D, RefList, 'ObjectOneOf'([Individual])).

% Disjunction
% E.g. John likes Mary or likes Bill.
condition_axiom(
	Or1 v Or2,
	RefList,
	'ClassAssertion'('ObjectUnionOf'([Or1Class, Or2Class]), Individual)
) :-
	condlist_and(D, Or1, RefList, Or1Class),
	condlist_and(D, Or2, RefList, Or2Class),
	is_toplevel(D, RefList, 'ObjectOneOf'([Individual])).

% object/8 where Count = "named"
% E.g. John ...
condition_axiom(
	object(_, ProperName, named, _, eq, 1)-_,
	_,
	'ClassAssertion'('OWLClass'(owl:'Thing'), Individual)
) :-
	get_namedindividual(ProperName, Individual).

% object/8 where Count = "dom" or "countable"
% E.g. man ...
% Common nouns are mapped to
% OWL anonymous individuals identified by Ref and typed by Name.
condition_axiom(
	object(Ref, Name, Count, _, QType, QNum)-_,
	_,
	'ClassAssertion'(NamedClass, 'AnonymousIndividual'(Ref))
) :-
	member(Count, [countable, dom]),
	member(QType, [eq, geq, na]),
	member(QNum, [1, na]),
	get_namedclass(Name, NamedClass).

% Property between an Individual and a Literal (DataProperty).
% E.g.: John's age is 30., John's address is "Poland".
condition_axiom(
	predicate(_, Predicate, Ref, DataItem)-_,
	RefList,
	'DataPropertyAssertion'(DataProperty, Name, '^^'(DataValue, DataTypeUri))
) :-
	is_toplevel(Ref, RefList, 'ObjectOneOf'([Name])),
	get_nameddataproperty(Predicate, DataProperty),
	dataitem_datavalue_datatypeuri(DataItem, DataValue, DataTypeUri).

% predicate(_, be, Ref1, Ref2)
%
% The copula-predicate maps to the SameIndividual-axiom.
%
% E.g. John is Mary.
condition_axiom(
	predicate(_, be, Ref1, Ref2)-_,
	RefList,
	'SameIndividual'([Name1, Name2])
) :-
	is_toplevel(Ref1, RefList, 'ObjectOneOf'([Name1])),
	is_toplevel(Ref2, RefList, 'ObjectOneOf'([Name2])).

% predicate(_, _, Ref1, Ref2)
%
% All the remaining predicates map to ObjectPropertyAssertion.
%
% E.g. John likes Mary.
%
% BUG: number is currently ignored, see e.g.: John owns more than 3 cars.
condition_axiom(
	predicate(_, Predicate, Ref1, Ref2)-_,
	RefList,
	'ObjectPropertyAssertion'(ObjectProperty, Name1, Name2)
) :-
	Predicate \= be,
	get_namedobjectproperty(Predicate, ObjectProperty),
	is_toplevel(Ref1, RefList, 'ObjectOneOf'([Name1])),
	is_toplevel(Ref2, RefList, 'ObjectOneOf'([Name2])).

% BUG: experimental: predicate(_, _, Ref1, Ref2)
% supports: John owns more than 3 cars.
% does not support: John owns more than 3 cars that Mary likes.
condition_axiom(
	predicate(_, Predicate, Ref1, Ref2)-_,
	RefList,
	'SubClassOf'('ObjectOneOf'([Name1]), PropertyRestriction)
) :-
	Predicate \= be,
	is_toplevel(Ref1, RefList, 'ObjectOneOf'([Name1])),
	is_toplevel(Ref2, RefList, genquant(_, Name, QType, QNum)),
	get_namedclass(Name, NamedClass),
	get_namedobjectproperty(Predicate, ObjectProperty),
	make_restr(QType, QNum, ObjectProperty, NamedClass, PropertyRestriction).

% Implication that maps to a SubObjectProperty-axiom
%
% E.g. If somebody X loves somebody Y then X likes Y.
%
% BUG: add: then-part contains inverse. (Not needed really, but maybe adds readability.)
%
% Note: the chain must not be empty!
condition_axiom(
	If => [predicate(_, Predicate, Ref1, RefN)-_],
	_,
	PropertyAxiom
) :-
	Predicate \= be,
	has_dom_for_member(Ref1, If, CondList1),
	is_chain(Ref1, RefN, CondList1, SubObjectPropertyChain),
	SubObjectPropertyChain = [_ | _],
	get_namedobjectproperty(Predicate, ObjectProperty),
	convert_to_owldl(
		'SubObjectPropertyOf'(
			'SubObjectPropertyChain'(SubObjectPropertyChain),
			ObjectProperty
		),
		PropertyAxiom
	).

% Implication that maps to a DisjointObjectProperties-axiom
%
% E.g. If somebody X loves somebody Y then it is false that X hates Y.
%
% Note: the chain must contain exactly 1 element.
condition_axiom(
	If => [-[predicate(_, Property2, Ref1, RefN)-_]],
	_,
	'DisjointObjectProperties'([Property1, ObjectProperty])
) :-
	Property2 \= be,
	has_dom_for_member(Ref1, If, CondList1),
	get_namedobjectproperty(Property2, ObjectProperty),
	is_chain(Ref1, RefN, CondList1, [Property1]).

% Implication that maps to a SubClassOf-axiom
%
% Examples:
%
% Every man is an animal and likes a dog and ...
% Every man who owns a car is a driver.
%
% This is by far the most complicated case to handle.
condition_axiom(
	If => Then,
	RefList,
	'SubClassOf'(IfClass, ThenClass)
) :-
	condlist_if(X, If, RefList, IfClass),
	condlist_and(X, Then, RefList, ThenClass).


%% is_chain(+Ref1:nvar, +RefN:nvar, +CondList:list, -SubObjectPropertyChain:list) is semidet.
%
% @param Ref1 is a discourse referent of the first object in the chain
% @param RefN is a discourse referent of the last object in the chain
% @param CondList is a list of DRS conditions
% @param SubObjectPropertyChain is a list of OWL property descriptions in the chain-order
%
is_chain(Ref, Ref, [], []).

is_chain(Ref1, RefN, CondList, [ObjectProperty | Chain]) :-
	select(predicate(_, Property, Ref1, Tmp)-_, CondList, CondList1),
	Property \= be,
	get_namedobjectproperty(Property, ObjectProperty),
	has_dom_for_member(Tmp, CondList1, CondList2),
	is_chain(Tmp, RefN, CondList2, Chain).

is_chain(Ref1, RefN, CondList, ['InverseObjectProperty'(ObjectProperty) | Chain]) :-
	select(predicate(_, Property, Tmp, Ref1)-_, CondList, CondList1),
	Property \= be,
	get_namedobjectproperty(Property, ObjectProperty),
	has_dom_for_member(Tmp, CondList1, CondList2),
	is_chain(Tmp, RefN, CondList2, Chain).


%% has_dom_for_member(+Ref:nvar, +CondListIn:list, -CondListOut:list) is nondet.
%
% @param Ref is a discourse referent
% @param CondListIn is a list of DRS conditions
% @param CondListOut is the remaining list of DRS conditions (after select/3)
%
% Given a discourse referent, selects its corresponding
% object condition, given that it corresponds to an indefinite pronoun
% ('somebody' or 'something') or the noun 'thing'.
%
has_dom_for_member(Ref, CondListIn, CondListOut) :-
	select(object(Ref, Name, Count, _, QType, QNum)-_, CondListIn, CondListOut),
	(
		Name = something, Count = dom
	;
		Name = somebody, Count = countable
	;
		Name = thing, Count = countable, QType = eq, QNum = 1
	).


%% convert_to_owldl(+Axiom1:term, -Axiom2:term) is semidet.
%
% @param Axiom1 OWL axiom
% @param Axiom2 Semantically equivalent OWL axiom
%
% Converts OWL 2 property axioms into a semantically equivalent yet
% syntactically simplified form, so that the resulting axiom is more
% backwards compatible with OWL DL. E.g. in case a property chain stands
% for a transitivity then we represent it with the TransitiveObjectProperty-axiom.
%
convert_to_owldl(
	'SubObjectPropertyOf'('SubObjectPropertyChain'(['InverseObjectProperty'(R)]), 'InverseObjectProperty'(S)),
	'SubObjectPropertyOf'(R, S)
) :- !.

convert_to_owldl(
	'SubObjectPropertyOf'('SubObjectPropertyChain'(['InverseObjectProperty'(R)]), 'ObjectProperty'(S)),
	'SubObjectPropertyOf'(R, 'InverseObjectProperty'('ObjectProperty'(S)))
) :- !.

convert_to_owldl(
	'SubObjectPropertyOf'('SubObjectPropertyChain'([R]), S),
	'SubObjectPropertyOf'(R, S)
) :- !.

convert_to_owldl(
	'SubObjectPropertyOf'('SubObjectPropertyChain'([R, R]), R),
	'TransitiveObjectProperty'(R)
) :- !.

convert_to_owldl(
	'SubObjectPropertyOf'('SubObjectPropertyChain'(SubObjectPropertyChain), Property),
	'SubObjectPropertyOf'('SubObjectPropertyChain'(SubObjectPropertyChain), Property)
).


%% condlist_if(+D:nvar, +CondList:list, +RefList:list, -If:term) is nondet.
%
% @param D is a distinguished discourse referent (the "subject" of an if-then sentence)
% @param CondList is a list of DRS conditions in the if-part
% @param RefList is a list of top-level discourse referents
% @param If is an OWL class description which is the left-side element of the resulting SubClassOf-axiom
%
% Example (proper name is distinguished): If there is a protein that activates Met then Met follows a gene.
%
condlist_if(D, CondListIn, RefList, If) :-
	select_object(D, CondListIn, RefList, NamedClass, QType, QNum, CondListOut),
	(
		QType = eq, QNum = 1
	;
		QType = na, QNum = na
	),
	condlist_classlist(D, CondListOut, RefList, DescriptionList),
	classlist_and([NamedClass | DescriptionList], If).


%% condlist_and(+D:nvar, +CondList:list, +RefList:list, -And:term) is nondet.
%
% @param D is a distinguished discourse referent (the "subject" of an if-then sentence)
% @param CondList is a list of DRS conditions (under negation, disjunction, or then-box)
% @param RefList is a list of top-level discourse referents
% @param And is an OWL class description (possibly a named class)
%
condlist_and(D, CondList, RefList, And) :-
	condlist_classlist(D, CondList, RefList, DescriptionList),
	classlist_and(DescriptionList, And).


%% classlist_and(+ClassList:list, -Class:term) is semidet.
%
% @param ClassList is a list of conjoined class descriptions
% @param Class is a class description
%
% The purpose of this predicate is to check whether the class-list
% contains just one element. In this case, it is not wrapped into the
% ObjectIntersectionOf class description, to keep things simple.
%
classlist_and([Description], Description).

classlist_and([D1, D2 | DescriptionList], 'ObjectIntersectionOf'([D1, D2 | DescriptionList])).


%% condlist_classlist(+D:nvar, +CondList:list, +RefList:list, -ClassList:list) is nondet.
%
% @param D is a discourse referent
% @param CondList is a list of DRS conditions
% @param RefList is a list of top-level discourse referents
% @param ClassList is a list of OWL class descriptions
%
% Map *all* the conditions in the list to corresponding
% OWL classes, i.e. no condition must be left unmapped.
%
condlist_classlist(_D, [], _RefList, []).

condlist_classlist(D, CondList, RefList, [And1Class | And2Class]) :-
	select(Condition, CondList, CondList1),
	condlist_class(D, Condition, CondList1, RefList, And1Class, CondList2),
	condlist_classlist(D, CondList2, RefList, And2Class).


%% condlist_classlist_err(+D:nvar, +CondList:list, +RefList:list) is nondet.
%
% @param D is a discourse referent
% @param CondList is a list of DRS conditions
% @param RefList is a list of top-level discourse referents
%
% These rules are here solely for error capture.
% BUG: This error capture is not complete, e.g. we can't see inside
% embedded DRSs.
%
condlist_classlist_err(_D, [], _RefList) :- !.

condlist_classlist_err(D, CondList, RefList) :-
	select(Condition, CondList, CondList1),
	condlist_class(D, Condition, CondList1, RefList, _, CondList2),
	!,
	condlist_classlist_err(D, CondList2, RefList).

% If condlist_class/5 failed then there must be a problem that
% we want to report.
condlist_classlist_err(_D, CondList, _RefList) :-
	illegal_conditions(CondList).



%% condlist_class(+D:nvar, Condition:term, +CondListIn:list, +RefList:list, +Class:term, -CondListOut:list) is nondet.
%
% @param D is a discourse referent
% @param Condition is a DRS condition
% @param CondListIn is a list of DRS conditions
% @param RefList is a list of top-level discourse referents
% @param Class is an OWL class description
% @param CondListOut is a list of remaining DRS conditions
%
% Note that we allow inverting the arguments of copula 'be', i.e.
% we consider the following equivalent:
%
% If a man likes somebody that is a person then the person owns a car.
% If a man likes a person that is a somebody then the person owns a car.
%
% If a man is John then the man is a person.
% If John is a man then the man is a person.
%
% This is similar to inverting the arguments of regular verbs, but while
% for regular verbs the Property changes into 'InverseObjectProperty'('ObjectProperty'(PropertyName)),
% the copula arguments are simply switched with no trace left behind.

% Copula ('be') predicate
% Every man is John.
% Every man is a human.
% Every man is something.
% All grass is some food.
% * Every man is at most 3 cars.
condlist_class(D, predicate(_, be, Subj, Obj)-_, CondList, RefList, EmbeddedClass, CondList2) :-
	(
		Subj = D, Obj = NewD
	;
		Obj = D, Subj = NewD
	),
	select_object(NewD, CondList, RefList, NamedClass, QType, QNum, CondList1),
	(
		QType = eq, QNum = 1
	;
		QType = na, QNum = na
	),
	follow_object(NewD, NamedClass, CondList1, RefList, EmbeddedClass, CondList2).

% Regular predicate with reflexive object
% Every man likes himself.
condlist_class(D, predicate(_, Property, D, D)-_, CondList, _RefList, 'ObjectExistsSelf'(ObjectProperty), CondList) :-
	Property \= be,
	get_namedobjectproperty(Property, ObjectProperty).

% Regular predicate with data object
% Every man's age is 20.
condlist_class(D, predicate(_, Property, D, DataItem)-_, CondList, _RefList, 'DataHasValue'(DataProperty, '^^'(DataValue, DataTypeUri)), CondList) :-
	Property \= be,
	get_nameddataproperty(Property, DataProperty),
	dataitem_datavalue_datatypeuri(DataItem, DataValue, DataTypeUri).

% Regular predicate with a dist. variable
% Every man likes a woman.
condlist_class(D, predicate(_, PropertyName, D, Obj)-_, CondList, RefList, Class, CondList2) :-
	PropertyName \= be,
	select_object(Obj, CondList, RefList, NamedClass, QType, QNum, CondList1),
	follow_object(Obj, NamedClass, CondList1, RefList, EmbeddedClass, CondList2),
	get_namedobjectproperty(PropertyName, ObjectProperty),
	make_restr(QType, QNum, ObjectProperty, EmbeddedClass, Class).

% Regular predicate with a dist. variable, inverted case
% Every man is liked by a woman.
condlist_class(D, predicate(_, PropertyName, Subj, D)-_, CondList, RefList, Class, CondList2) :-
	PropertyName \= be,
	select_object(Subj, CondList, RefList, NamedClass, QType, QNum, CondList1),
	follow_object(Subj, NamedClass, CondList1, RefList, EmbeddedClass, CondList2),
	get_namedobjectproperty(PropertyName, ObjectProperty),
	make_restr(QType, QNum, 'InverseObjectProperty'(ObjectProperty), EmbeddedClass, Class).

% Negation
% Every man is not a table.
condlist_class(D, -Not, CondList, RefList, 'ObjectComplementOf'(NotClass), CondList) :-
	condlist_and(D, Not, RefList, NotClass).

% Disjunction
% Every man is policeman or is not a table.
condlist_class(D, Or1 v Or2, CondList, RefList, 'ObjectUnionOf'([Or1Class, Or2Class]), CondList) :-
	condlist_and(D, Or1, RefList, Or1Class),
	condlist_and(D, Or2, RefList, Or2Class).


%% select_object(+D:nvar, +CondListIn:list, +RefList:list, -NamedClass:term, -QType:atom, -QNum:atomic, -CondListOut:list) is nondet.
%
% @param D is a distinguished discourse referent
% @param CondListIn is a list of DRS conditions
% @param RefList is a list of top-level discourse referents
% @param NamedClass is either OWLClass or ObjectOneOf([_])
% @param QType is in {na, eq, geq, leq, greater, less, exactly}
% @param QNum is a positive integer or 'na' (not available)
% @param CondListOut is a list of remaining DRS conditions
%
select_object(D, CondList, RefList, 'ObjectOneOf'([Individual]), eq, 1, CondList) :-
	is_toplevel(D, RefList, 'ObjectOneOf'([Individual])).

select_object(D, CondListIn, _RefList, NamedClass, QType, QNum, CondListOut) :-
	select(object(D, Name, _, _, QType, QNum)-_, CondListIn, CondListOut),
	get_namedclass(Name, NamedClass).


%% follow_object(+D:nvar, +NamedClass:atom, +CondListIn:list, +RefList:list, -Class:term, -CondListOut:list) is semidet.
%
% @param D is a distinguished discourse referent
% @param NamedClass is either OWLClass or ObjectOneOf([_])
% @param CondListIn is a list of DRS conditions
% @param RefList is a list of top-level discourse referents
% @param Class is an OWL class description
% @param CondListOut is a list of remaining DRS conditions
%
% We first try to build a complete class-list (and consume all the conditions).
% 1. If this succeeds then the returned class-list is either empty or not.
% 1.1. In case of an empty class-list, Class is simply a named class.
% 1.2. In case of a non-empty class-list, Class is an intersection with the named class as the first element.
%
% 2. If building the complete class-list fails (i.e. some conditions are not consumed)
% then we try to build just one class description.
% 2.1 If this succeeds then we might be able to build some more, i.e. we call follow_object/6 recursively
% 2.1.1 If building some more fails then we return the remaining conditions.
% 2.2 If building just one class description fails, then we return the named class.
%
% BUG: Don't produce so many owl:Things (see: follow_object(D, 'OWLClass'(owl:'Thing'), ...))
%
follow_object(D, NamedClass, CondList, RefList, Class, CondListOut) :-
	(
		condlist_classlist(D, CondList, RefList, ClassList)
	->
		(
			ClassList = []
		->
			Class = NamedClass
		;
			Class = 'ObjectIntersectionOf'([NamedClass | ClassList])
		),
		CondListOut = []
	;
		(
			(
				select(Condition, CondList, CondList1),
				condlist_class(D, Condition, CondList1, RefList, Class1, CondList2)
			)
		->
			(
				follow_object(D, 'OWLClass'(owl:'Thing'), CondList2, RefList, Class2, CondList3)
			->
				Class = 'ObjectIntersectionOf'([NamedClass, Class1, Class2]),
				CondListOut = CondList3
			;
				Class = 'ObjectIntersectionOf'([NamedClass, Class1]),
				CondListOut = CondList2
			)
		;
			Class = NamedClass,
			CondListOut = CondList
		)
	).


%% is_toplevel(+Ref:nvar, +RefList:list, -ObjectOneOf:term) is nondet.
%
% @param Ref is a discourse referent
% @param RefList is a list of top-level discourse referents
% @param ObjectOneOf is an OWL ObjectOneOf-class
%
% Succeeds if Ref is among the top-level referents (i.e. it corresponds to
% a proper name or to a common noun that is not under negation, disjunction,
% or implication).
%
is_toplevel(Ref, RefList, ObjectOneOf) :-
	member(ref_oneof(Ref, ObjectOneOf), RefList).


%% make_restr(+QType:atom, +QNum:atomic, +Property:term, +Class:term, -Restriction:term) is semidet.
%
% @param QType is in {na, eq, geq, leq, greater, less, exactly}
% @param QNum is a positive integer or 'na' (not available)
% @param Property is an OWL property description
% @param Class is an OWL class description
% @param Restriction is an OWL class description built from QType, QNum, Property, and Class
%
make_restr(na, na, Property, Class, 'ObjectSomeValuesFrom'(Property, Class)).
make_restr(eq, 1, Property, Class, 'ObjectSomeValuesFrom'(Property, Class)).
make_restr(geq, 1, Property, Class, 'ObjectSomeValuesFrom'(Property, Class)).
make_restr(eq, QNum, Property, Class, 'ObjectMinCardinality'(QNum, Property, Class)) :-
	QNum > 1.
make_restr(geq, QNum, Property, Class, 'ObjectMinCardinality'(QNum, Property, Class)) :-
	QNum > 1.
make_restr(leq, QNum, Property, Class, 'ObjectMaxCardinality'(QNum, Property, Class)).
make_restr(exactly, QNum, Property, Class, 'ObjectExactCardinality'(QNum, Property, Class)).
make_restr(less, QNum, Property, Class, 'ObjectMaxCardinality'(Num, Property, Class)) :-
	Num is QNum - 1.
make_restr(greater, QNum, Property, Class, 'ObjectMinCardinality'(Num, Property, Class)) :-
	Num is QNum + 1.


%% get_namedclass(+Name:atom, +NS:atom, -OWLClass:term) is semidet.
%% get_namedclass(+Name:atom, -OWLClass:term) is semidet.
%
% @param Name is the name of the class
% @param NS is a namespace identifier
% @param OWLClass is a named OWL class (possibly owl:Thing)
%
get_namedclass(Name, OWLClass) :-
	get_namedclass(Name, '', OWLClass).

get_namedclass(somebody, _, 'OWLClass'(owl:'Thing')) :- !.
get_namedclass(something, _, 'OWLClass'(owl:'Thing')) :- !.
get_namedclass(thing, _, 'OWLClass'(owl:'Thing')) :- !.
get_namedclass(Name, NS, 'OWLClass'(NS:Name)).


%% get_namedobjectproperty(+Name:atom, +NS:atom, -ObjectProperty:term) is semidet.
%% get_namedobjectproperty(+Name:atom, -ObjectProperty:term) is semidet.
%
% @param Name is the name of the object property
% @param NS is a namespace identifier
% @param ObjectProperty is a named object property
%
get_namedobjectproperty(Name, ObjectProperty) :-
	get_namedobjectproperty(Name, '', ObjectProperty).

get_namedobjectproperty(Name, NS, 'ObjectProperty'(NS:Name)).


%% get_nameddataproperty(+Name:atom, +NS:atom, -DataProperty:term) is semidet.
%% get_nameddataproperty(+Name:atom, -DataProperty:term) is semidet.
%
% @param Name is the name of the data property
% @param NS is a namespace identifier
% @param DataProperty is a named data property
%
get_nameddataproperty(Name, DataProperty) :-
	get_nameddataproperty(Name, '', DataProperty).

get_nameddataproperty(Name, NS, 'DataProperty'(NS:Name)).


%% get_namedindividual(+Name:atom, +NS:atom, -Individual:term) is semidet.
%% get_namedindividual(+Name:atom, -Individual:term) is semidet.
%
% @param Name is the name of the individual
% @param NS is a namespace identifier
% @param Individual is a named individual
%
get_namedindividual(Name, Individual) :-
	get_namedindividual(Name, '', Individual).

get_namedindividual(Name, NS, 'Individual'(NS:Name)).


%% datatype_datatypeuri(+DataType:atom, -DataTypeUri:atom) is semidet.
%
% @param DataType is a DRS datatype, currently in {integer, string, real}
% @param DataTypeUri is the corresponding XMLSchema datatype URI
%
datatype_datatypeuri(integer, 'http://www.w3.org/2001/XMLSchema#integer').
datatype_datatypeuri(string, 'http://www.w3.org/2001/XMLSchema#string').
datatype_datatypeuri(real, 'http://www.w3.org/2001/XMLSchema#double').


%% dataitem_datavalue_datatypeuri(+DataItem:term, -DataValue:term, -DataTypeUri:atom) is semidet.
%
% @param DataItem is an ACE data-item (e.g. int(10), real(3.14), string('Go!'))
% @param DataValue is an ACE data-value (e.g. 10, 3.14, 'Go!')
% @param DataTypeUri is the corresponding XMLSchema datatype URI
%
dataitem_datavalue_datatypeuri(string(DataValue), DataValue, DataTypeUri) :-
	datatype_datatypeuri(string, DataTypeUri).

dataitem_datavalue_datatypeuri(int(DataItem), DataItem, DataTypeUri) :-
	datatype_datatypeuri(integer, DataTypeUri).

dataitem_datavalue_datatypeuri(real(DataItem), DataItem, DataTypeUri) :-
	datatype_datatypeuri(real, DataTypeUri).


%% ontology_url(-UrlPrefix:atom) is det.
%
% @param UrlPrefix is the prefix for ACE words used as OWL names
%
ontology_url('http://attempto.ifi.uzh.ch/ontologies/owlswrl/test').
