% This file is part of the Attempto Parsing Engine (APE).
% Copyright 2008, Kaarel Kaljurand <kaljurand@gmail.com>.
%
% The Attempto Parsing Engine (APE) is free software: you can redistribute it and/or modify it
% under the terms of the GNU Lesser General Public License as published by the Free Software
% Foundation, either version 3 of the License, or (at your option) any later version.
%
% The Attempto Parsing Engine (APE) is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
% PURPOSE. See the GNU Lesser General Public License for more details.
%
% You should have received a copy of the GNU Lesser General Public License along with the Attempto
% Parsing Engine (APE). If not, see http://www.gnu.org/licenses/.


:- module(owlfss_owlxml, [
		owlfss_owlxml/2
	]).


/** <module> OWL 2/SWRL Functional-Style Syntax to OWL 2 XML converter

@author Kaarel Kaljurand
@version 2008-03-31

@bug support for datatype properties is not implemented or buggy
@tbd support for SWRL is missing (there is no official standard for this)
@tbd anonymous individuals should not be turned into regular individuals
@tbd what is correct, SameIndividual or SameIndividuals?

*/


%% owlfss_owlxml(+OWL:term, -XML:term) is det.
%
% @param OWL is OWL ontology in Functional-Style Syntax (Prolog notation)
% @param XML is OWL ontology in XML (SWI-Prolog's XML notation)
%
% @bug: use ontologyURI instead of URI
%
owlfss_owlxml('Ontology'(Name, AxiomList), XML) :-
	ellist_termlist(ElList, Name, AxiomList),

	XML = element('Ontology', [
		'xml:base'='http://www.w3.org/2006/12/owl11-xml#',
		xmlns='http://www.w3.org/2006/12/owl11-xml#',
		'URI'=Name
		], ElList).


%% ellist_termlist(+ElementList:list, -TermList:list) is det.
%
% @param ElementList is a list of XML elements
% @param TermList is a list of Prolog terms (representing OWL expressions)
%
ellist_termlist([], _, []).

ellist_termlist([El | ElList], Prefix, [Term | TermList]) :-
	el_term(El, Prefix, Term),
	!,
	ellist_termlist(ElList, Prefix, TermList).


%% el_term(+Element:term, +OWLEntity:term) is det.
%
% @param Element is an XML element
% @param OWLEntity is an OWL entity
%
el_term(element('Constant', [], []), _, '^^'('', '')) :- !.

el_term(element('Constant', [], [PCDATA]), _, '^^'(PCDATA, '')) :- !.

el_term(element('Constant', ['datatypeURI' = Datatype], [PCDATA]), _, '^^'(Data, Datatype)) :-
	!,
	datatype_pcdata_data(PCDATA, Datatype, Data).

el_term(element('Constant', ['datatypeURI' = Datatype], []), _, '^^'('', Datatype)) :- !.

el_term(element(Term, [], ElList), Prefix, Return) :-
	is_owl_list(Term, TermList, Return),
	!,
	ellist_termlist(ElList, Prefix, TermList).

el_term(element(Term, [cardinality = Cardinality], ElList), Prefix, Return) :-
	Return =.. [Term, Cardinality | TermList],
	number(Cardinality),
	!,
	ellist_termlist(ElList, Prefix, TermList).

/*
el_term(element(Term, [facet = Facet], ElList), Return) :-
	!,
	ellist_termlist(ElList, TermList),
	Return =.. [Term, Facet | TermList].
*/

el_term(element(Term, ['URI' = Uri], []), DefaultPrefix, Base) :-
	is_base(Term, NS, Name, Base),
	uri_name(Uri, NS, DefaultPrefix, Name),
	!.

el_term(element(Term, [], ElList), Prefix, Return) :-
	!,
	Return =.. [Term | TermList],
	ellist_termlist(ElList, Prefix, TermList).

/*
el_term(element(Term, ['annotationURI' = Uri], ElList), Return) :-
	Return =.. [Term, Uri, Term1],
	ellist_termlist(ElList, [Term1]).
*/

el_term(PCDATA, _, PCDATA) :-
	atom(PCDATA).


%% is_base(+OWLEntityName:atom, +NS:atom, +URIName:atom, -OWLEntity:term) is det.
%
% @bug Datatype should also have a namespace identifier
%
is_base('OWLClass', NS, Name, 'OWLClass'(NS:Name)).
is_base('ObjectProperty', NS, Name, 'ObjectProperty'(NS:Name)).
is_base('DataProperty', NS, Name, 'DataProperty'(NS:Name)).
is_base('Individual', NS, Name, 'Individual'(NS:Name)).
is_base('Individual', '', IName, 'AnonymousIndividual'('$VAR'(Name))) :- atom_concat('I', Name, IName).
is_base('Datatype', '', Name, 'Datatype'(Name)).


%% is_owl_list(+ElementName:atom) is det.
%
% @param ElementName is an OWL XML element name
%
% Succeeds if the element denoted by the name has a set/list type
% in OWL structural specification.
%
% TODO:
%
% is_owl_list('SameIndividual', TermList, 'SameIndividual'(TermList)).
%
% EntityAnnotation
% DataAllValuesFrom
% DataSomeValuesFrom
% DisjointUnion
%
is_owl_list('DataOneOf', TermList, 'DataOneOf'(TermList)).
is_owl_list('ObjectUnionOf', TermList, 'ObjectUnionOf'(TermList)).
is_owl_list('ObjectIntersectionOf', TermList, 'ObjectIntersectionOf'(TermList)).
is_owl_list('ObjectOneOf', TermList, 'ObjectOneOf'(TermList)).
is_owl_list('EquivalentClasses', TermList, 'EquivalentClasses'(TermList)).
is_owl_list('DisjointClasses', TermList, 'DisjointClasses'(TermList)).
is_owl_list('SubObjectPropertyChain', TermList, 'SubObjectPropertyChain'(TermList)).
is_owl_list('EquivalentObjectProperties', TermList, 'EquivalentObjectProperties'(TermList)).
is_owl_list('DisjointObjectProperties', TermList, 'DisjointObjectProperties'(TermList)).
is_owl_list('EquivalentDataProperties', TermList, 'EquivalentDataProperties'(TermList)).
is_owl_list('DisjointDataProperties', TermList, 'DisjointDataProperties'(TermList)).
is_owl_list('SameIndividuals', TermList, 'SameIndividual'(TermList)).
is_owl_list('SameIndividuals', TermList, 'SameIndividuals'(TermList)).
is_owl_list('DifferentIndividuals', TermList, 'DifferentIndividuals'(TermList)).

is_owl_list('DatatypeRestriction', TermList, 'DatatypeRestriction'(TermList)).


%% uri_name(+URI:atom, +NS:atom, +Prefix:atom, -Name:term) is det.
%
%
uri_name('http://www.w3.org/2002/07/owl#Thing', owl, _, 'Thing') :- !.
uri_name('http://www.w3.org/2002/07/owl#Nothing', owl, _, 'Nothing') :- !.

uri_name('http://attempto.ifi.uzh.ch/ace#Superman', ace, _, 'Superman') :- !.
uri_name('http://attempto.ifi.uzh.ch/ace#know', ace, _, 'know') :- !.

uri_name(Uri, '', Prefix, Name) :-
	!,
	concat_atom([Prefix, Name], '#', Uri).

uri_name(Uri, NS, _, Name) :-
	concat_atom([NS, Name], '#', Uri).


%% datatype_pcdata_data(+PCDATA:atom, +Datatype:atom, -Data:atomic) is det.
%
% @param PCDATA is a data value (atom)
% @param Datatype is an XML Schema datatype
% @param Data is a data value (atom or number)
%
% The purpose of this rule is to convert come of the PCDATA
% into Prolog numbers. In case of failure, the PCDATA is returned
% as it it, i.e. it will be represented as an atom.
%
% Note that if the PCDATA (i.e. an atom) cannot be converted
% into a number then atom_number/2 throws an exception.
% This is caught at higher level (for the time being).
%
datatype_pcdata_data(PCDATA, Datatype, Data) :-
	is_numbertype(Datatype),
	!,
	atom_number(PCDATA, Data).

datatype_pcdata_data(PCDATA, _, PCDATA).


%% is_numbertype(+Datatype:atom) is det.
%
% @param Datatype is an XML Schema datatype
%
is_numbertype('http://www.w3.org/2001/XMLSchema#int').
is_numbertype('http://www.w3.org/2001/XMLSchema#integer').
is_numbertype('http://www.w3.org/2001/XMLSchema#nonNegativeInteger').
is_numbertype('http://www.w3.org/2001/XMLSchema#float').
is_numbertype('http://www.w3.org/2001/XMLSchema#double').
is_numbertype('http://www.w3.org/2001/XMLSchema#short').
is_numbertype('http://www.w3.org/2001/XMLSchema#long').
