% This file is part of the Attempto Parsing Engine (APE).
% Copyright 2008, Attempto Group, University of Zurich (see http://attempto.ifi.uzh.ch).
%
% The Attempto Parsing Engine (APE) is free software: you can redistribute it and/or modify it
% under the terms of the GNU Lesser General Public License as published by the Free Software
% Foundation, either version 3 of the License, or (at your option) any later version.
%
% The Attempto Parsing Engine (APE) is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
% PURPOSE. See the GNU Lesser General Public License for more details.
%
% You should have received a copy of the GNU Lesser General Public License along with the Attempto
% Parsing Engine (APE). If not, see http://www.gnu.org/licenses/.


:- module(typeddrs_to_untypeddrs, [typeddrs_to_untypeddrs/2]).

/** <module> Typed DRS to Untyped DRS converter

Removes
argument positions 7 and 8 from the object/8 conditions,
the last argument position from the predicate/[4,5,6] conditions,
the argument position 4 from the modifier/4 conditions, and
the last argument position from the property/[4,5,7] conditions.

@author Kaarel Kaljurand
@version 2007-05-02

*/


% Operators used in the DRS.
:- op(400, fx, -).
:- op(400, fx, ~).
:- op(400, fx, <>).
:- op(400, fx, []).
:- op(500, xfx, =>).
:- op(500, xfx, v).


%% typeddrs_to_untypeddrs(+TypedDrs, -UntypedDrs) is det.
%% typeddrs_to_untypeddrs_x(+TypedDrs, -UntypedDrs) is det.
%
%
typeddrs_to_untypeddrs(OldDrs, NewDrs) :-
	typeddrs_to_untypeddrs_x(OldDrs, NewDrs).

typeddrs_to_untypeddrs_x(drs(Dom, TCondList), drs(Dom, UCondList)) :-
	typedcondlist_untypedcondlist(TCondList, UCondList).


%% typedcondlist_untypedcondlist(+TypedConditionList:list, -UntypedConditionList:list) is det.
%
%
typedcondlist_untypedcondlist([], []).

typedcondlist_untypedcondlist([TCond | TCondList], [UCond | UCondList]) :-
	typedcond_untypedcond(TCond, UCond),
	typedcondlist_untypedcondlist(TCondList, UCondList).


%% typedcond_untypedcond(+TypedCondition:term, -UntypedCondition:term) is det.
%
%
typedcond_untypedcond(-TDrs, -UDrs) :-
	!,
	typeddrs_to_untypeddrs_x(TDrs, UDrs).

typedcond_untypedcond(~TDrs, ~UDrs) :-
	!,
	typeddrs_to_untypeddrs_x(TDrs, UDrs).

typedcond_untypedcond(<>TDrs, <>UDrs) :-
	!,
	typeddrs_to_untypeddrs_x(TDrs, UDrs).

typedcond_untypedcond([]TDrs, []UDrs) :-
	!,
	typeddrs_to_untypeddrs_x(TDrs, UDrs).

typedcond_untypedcond(Label:TDrs, Label:UDrs) :-
	!,
	typeddrs_to_untypeddrs_x(TDrs, UDrs).

typedcond_untypedcond(TDrs1 => TDrs2, UDrs1 => UDrs2) :-
	!,
	typeddrs_to_untypeddrs_x(TDrs1, UDrs1),
	typeddrs_to_untypeddrs_x(TDrs2, UDrs2).

typedcond_untypedcond(TDrs1 v TDrs2, UDrs1 v UDrs2) :-
	!,
	typeddrs_to_untypeddrs_x(TDrs1, UDrs1),
	typeddrs_to_untypeddrs_x(TDrs2, UDrs2).

typedcond_untypedcond(object(X1, X2, X3, X4, X5, X6, _X7, _X8)-N, object(X1, X2, X3, X4, X5, X6)-N) :- !.

typedcond_untypedcond(predicate(X1, X2, X3, _X4)-N, predicate(X1, X2, X3)-N) :- !.

typedcond_untypedcond(predicate(X1, X2, X3, X4, _X5)-N, predicate(X1, X2, X3, X4)-N) :- !.

typedcond_untypedcond(predicate(X1, X2, X3, X4, X5, _X6)-N, predicate(X1, X2, X3, X4, X5)-N) :- !.

typedcond_untypedcond(modifier_adv(X1, X2, X3, _X4)-N, modifier_adv(X1, X2, X3)-N) :- !.

typedcond_untypedcond(modifier_pp(X1, X2, X3, _X4)-N, modifier_pp(X1, X2, X3)-N) :- !.

typedcond_untypedcond(property(X1, X2, X3, _X4)-N, property(X1, X2, X3)-N) :- !.

typedcond_untypedcond(property(X1, X2, X3, X4, _X5)-N, property(X1, X2, X3, X4)-N) :- !.

typedcond_untypedcond(property(X1, X2, X3, X4, X5, X6, _X7)-N, property(X1, X2, X3, X4, X5, X6)-N) :- !.

typedcond_untypedcond(Condition, Condition).
